#      Author: Carlos Olivares
#
#   Copyright (C) 2020 Carlos Olivares
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.


Igor_event_type_list = ['GeneChoice', 'Deletion', 'Insertion', 'DinucMarkov']

Igor_seq_type_list = ["V_gene", "VD_genes", "D_gene", "DJ_gene", "VJ_gene", "J_gene", "VDJ_genes", "Undefined_gene"]

Igor_seq_side_list = ["Five_prime", "Three_prime", "Undefined_side"]

Igor_nickname_list = ["v_choice", "j_choice", "d_gene", "v_3_del",
                     "d_3_del", "d_5_del", "j_5_del",
                     "vd_ins", "vd_dinucl", "dj_ins", "dj_dinucl",
                     "vj_ins", "vj_dinucl"]


IgorRec_Event_default_dict = {
    # GeneChoice;V_gene;Undefined_side;7;v_choice
    'v_choice': {
        "event_type": "GeneChoice",
        "seq_type": "V_gene",
        "seq_side": "Undefined_side",
        "priority": 7,
        "realizations": list(),
        "name": "",
        "nickname": "v_choice"
    },
    # GeneChoice;J_gene;Undefined_side;7;j_choice
    'j_choice': {
        "event_type": "GeneChoice",
        "seq_type": "J_gene",
        "seq_side": "Undefined_side",
        "priority": 7,
        "realizations": list(),
        "name": "",
        "nickname": "j_choice"
    },
    # GeneChoice;D_gene;Undefined_side;6;d_gene
    'd_gene': {
        "event_type": "GeneChoice",
        "seq_type": "D_gene",
        "seq_side": "Undefined_side",
        "priority": 6,
        "realizations": list(),
        "name": "",
        "nickname": "d_gene"
    },
    # Deletion;V_gene;Three_prime;5;v_3_del
    'v_3_del': {
        "event_type": "Deletion",
        "seq_type": "V_gene",
        "seq_side": "Three_prime",
        "priority": 5,
        "realizations": list(),
        "name": "",
        "nickname": "v_3_del"
    },
    # Deletion;D_gene;Three_prime;5;d_3_del
    'd_3_del': {
        "event_type": "Deletion",
        "seq_type": "D_gene",
        "seq_side": "Three_prime",
        "priority": 5,
        "realizations": list(),
        "name": "",
        "nickname": "d_3_del"
    },
    # Deletion;D_gene;Five_prime;5;d_5_del
    'd_5_del': {
        "event_type": "Deletion",
        "seq_type": "D_gene",
        "seq_side": "Five_prime",
        "priority": 5,
        "realizations": list(),
        "name": "",
        "nickname": "d_5_del"
    },
    # Deletion;J_gene;Five_prime;5;j_5_del
    'j_5_del': {
        "event_type": "Deletion",
        "seq_type": "J_gene",
        "seq_side": "Five_prime",
        "priority": 5,
        "realizations": list(),
        "name": "",
        "nickname": "j_5_del"
    },
    # Insertion;VD_genes;Undefined_side;4;vd_ins
    'vd_ins': {
        "event_type": "Insertion",
        "seq_type": "VD_genes",
        "seq_side": "Undefined_side",
        "priority": 4,
        "realizations": list(),
        "name": "",
        "nickname": "vd_ins"
    },
    # DinucMarkov;VD_genes;Undefined_side;3;vd_dinucl
    'vd_dinucl': {
        "event_type": "DinucMarkov",
        "seq_type": "VD_genes",
        "seq_side": "Undefined_side",
        "priority": 3,
        "realizations": list(),
        "name": "",
        "nickname": "vd_dinucl"
    },
    # Insertion;DJ_gene;Undefined_side;2;dj_ins
    'dj_ins': {
        "event_type": "Insertion",
        "seq_type": "DJ_gene",
        "seq_side": "Undefined_side",
        "priority": 2,
        "realizations": list(),
        "name": "",
        "nickname": "dj_ins"
    },
    # DinucMarkov;DJ_gene;Undefined_side;1;dj_dinucl
    'dj_dinucl': {
        "event_type": "DinucMarkov",
        "seq_type": "DJ_gene",
        "seq_side": "Undefined_side",
        "priority": 1,
        "realizations": list(),
        "name": "",
        "nickname": "dj_dinucl"
    },
    #Insertion;VJ_gene;Undefined_side;4;vj_ins
    'vj_ins': {
        "event_type": "Insertion",
        "seq_type": "VJ_gene",
        "seq_side": "Undefined_side",
        "priority": 4,
        "realizations": list(),
        "name": "",
        "nickname": "vj_ins"
    },
    #DinucMarkov;VJ_gene;Undefined_side;3;vj_dinucl
    'vj_dinucl': {
        "event_type": "DinucMarkov",
        "seq_type": "VJ_gene",
        "seq_side": "Undefined_side",
        "priority": 3,
        "realizations": list(),
        "name": "",
        "nickname": "vj_dinucl"
    }
}
#SingleErrorRate


Igor_VDJ_default_nickname_list = ["v_choice", "j_choice", "d_gene", "v_3_del",
                     "d_3_del", "d_5_del", "j_5_del",
                     "vd_ins", "vd_dinucl", "dj_ins", "dj_dinucl"]

Igor_VJ_default_nickname_list = ["v_choice", "j_choice",
                                 "v_3_del", "j_5_del",
                                 "vj_ins", "vj_dinucl"]

Igor_VDJ_default_parents_dict = {
    'v_choice' : [],
    'j_choice' : ['v_choice'],
    'd_gene' : ['v_choice', 'j_choice'],
    'v_3_del' : ['v_choice'],
    'd_5_del' : ['d_gene'],
    'd_3_del' : ['d_gene', 'd_5_del'],
    'j_5_del' : ['j_choice'],
    'vd_ins' : [],
    'vd_dinucl' : [],
    'dj_ins' : [],
    'dj_dinucl' : []
}

Igor_VJ_default_parents_dict = {
    'v_choice' : [],
    'j_choice' : ['v_choice'],
    'v_3_del' : ['v_choice'],
    'j_5_del' : ['j_choice'],
    'vj_ins' : [],
    'vj_dinucl' : []
}

Igor_to_AIRR_dict={
    "seq_index" : "sequence_id",
    "sequence" : "sequence",
    "v_choice" : "v_call",
    "d_gene" : "d_call",
    "j_choice" : "j_call" #,
    # "np1",
    # "np2"
}



# SQL commands to create database
sqlcmd_ct = dict()

sqlcmd_ct['indexed_sequences'] = """
-- IgorIndexedSeq table
CREATE TABLE IF NOT EXISTS IgorIndexedSeq (
    seq_index integer PRIMARY KEY,
    sequence text NOT NULL
);
"""

sqlcmd_ct['indexed_CDR3'] = """
-- IgorIndexedSeq table
CREATE TABLE IF NOT EXISTS IgorIndexedCDR3 (
    seq_index integer PRIMARY KEY,
    v_anchor integer,
    j_anchor integer,
    CDR3 text,
    CDR3_aa text
);
"""
# """
# -- IgorIndexedSeq table
# CREATE TABLE IF NOT EXISTS IgorIndexedCDR3 (
#     seq_index integer PRIMARY KEY,
#     v_anchor integer,
#     j_anchor integer,
#     CDR3nt text,
#     CDR3aa text
# );
# """

sqlcmd_ct['genomicVs'] = """
-- IgorVGeneTemplate table
CREATE TABLE IF NOT EXISTS IgorVGeneTemplate (
    vgene_id integer PRIMARY KEY,
    gene_name text NOT NULL,
    sequence text NOT NULL
);
"""

sqlcmd_ct['genomicDs'] = """
-- IgorDGeneTemplate table
CREATE TABLE IF NOT EXISTS IgorDGeneTemplate (
    dgene_id integer PRIMARY KEY,
    gene_name text NOT NULL,
    sequence text NOT NULL
);
"""

sqlcmd_ct['genomicJs'] = """
-- IgorJGeneTemplate table
CREATE TABLE IF NOT EXISTS IgorJGeneTemplate (
    jgene_id integer PRIMARY KEY,
    gene_name text NOT NULL,
    sequence text NOT NULL
);
"""

sqlcmd_ct['geneVCDR3Anchors'] = """
-- IgorVGeneCDR3Anchors table
CREATE TABLE IF NOT EXISTS IgorVGeneCDR3Anchors (
    vgene_id integer,
    anchor_index integer,
    function text,
    FOREIGN KEY (vgene_id)  REFERENCES IgorVGeneTemplate(vgene_id)
);
"""

sqlcmd_ct['geneJCDR3Anchors'] = """
-- IgorJGeneCDR3Anchors table
CREATE TABLE IF NOT EXISTS IgorJGeneCDR3Anchors (
    jgene_id integer,
    anchor_index integer,
    function text,
    FOREIGN KEY (jgene_id)  REFERENCES IgorJGeneTemplate(jgene_id)
);
"""

sqlcmd_ct['V_alignments'] = """
-- IgorVAlignments table
CREATE TABLE IF NOT EXISTS IgorVAlignments (
    seq_index integer NOT NULL,
    vgene_id integer NOT NULL,
    score integer,
    offset integer,
    insertions text NOT NULL,
    deletions  text NOT NULL,
    mismatches text NOT NULL,
    length integer,
    offset_5_p_align integer,
    offset_3_p_align integer,
    FOREIGN KEY (seq_index) REFERENCES IgorIndexedSeq   (seq_index),
    FOREIGN KEY (vgene_id)  REFERENCES IgorVGeneTemplate(vgene_id)
    --PRIMARY KEY (seq_index, vgene_id)
);
"""

sqlcmd_ct['D_alignments'] = """
-- IgorDAlignments table
CREATE TABLE IF NOT EXISTS IgorDAlignments (
    seq_index integer NOT NULL,
    dgene_id integer NOT NULL,
    score integer,
    offset integer,
    insertions text NOT NULL,
    deletions  text NOT NULL,
    mismatches text NOT NULL,
    length integer,
    offset_5_p_align integer,
    offset_3_p_align integer,
    FOREIGN KEY (seq_index) REFERENCES IgorIndexedSeq   (seq_index),
    FOREIGN KEY (dgene_id)  REFERENCES IgorDGeneTemplate(dgene_id)
    --PRIMARY KEY (seq_index, dgene_id)
);
"""

sqlcmd_ct['J_alignments'] = """
-- IgorJAlignments table
CREATE TABLE IF NOT EXISTS IgorJAlignments (
    seq_index integer NOT NULL,
    jgene_id integer NOT NULL,
    score integer,
    offset integer,
    insertions text NOT NULL,
    deletions  text NOT NULL,
    mismatches text NOT NULL,
    length integer,
    offset_5_p_align integer,
    offset_3_p_align integer,
    FOREIGN KEY (seq_index) REFERENCES IgorIndexedSeq   (seq_index),
    FOREIGN KEY (jgene_id)  REFERENCES IgorJGeneTemplate(jgene_id)
    --PRIMARY KEY (seq_index, jgene_id)
);
"""


############# IGOR MODEL TABLES #############
sqlcmd_ct['MP_Event_list'] = """
-- IgorMP_Event_list table
--     event_id integer,
CREATE TABLE IF NOT EXISTS IgorMP_Event_list (
    nickname text NOT NULL PRIMARY KEY,
    event_type text,
    seq_type text,
    seq_side text,
    priority integer,
    realizations_table text,
    name text
);
"""

sqlcmd_ct['ER_event_template'] = """
-- ER_event_template table
CREATE TABLE IF NOT EXISTS {} (
    id integer NOT NULL PRIMARY KEY,
    value text,
    name text
);
"""

sqlcmd_ct['MP_Edges'] = """
-- IgorMP_Edges table
CREATE TABLE IF NOT EXISTS IgorMP_Edges (
    parent_event text NOT NULL,
    child_event text NOT NULL,
    FOREIGN KEY (parent_event) REFERENCES IgorMP_Event_list (nickname),
    FOREIGN KEY (child_event)  REFERENCES IgorMP_Event_list (nickname)
);
"""

sqlcmd_ct['MP_ErrorRate'] = """
-- IgorMP_ErrorRate table
CREATE TABLE IF NOT EXISTS IgorMP_ErrorRate (
    error_type text NOT NULL,
    error_values text NOT NULL
);
"""

def sqlcmd_ct_Model_Marginals(event_nickname, list_dependencies:list):
    # from the xarray get the list of events something like:
    lista = [event_nickname]+list_dependencies
    str_column_ct = "id_{} integer NOT NULL" #.format(event_nickname)
    str_foreign_key_ct = "FOREIGN KEY (id_{}) REFERENCES IgorER_{} (id)" #.format(event_nickname)

    sqlcmd_table_fields_ct = ",\n".join([str_column_ct.format(evento_nickname) for evento_nickname in lista])
    sqlcmd_foreign_keys_ct = ",\n".join([str_foreign_key_ct.format(evento_nickname, evento_nickname) for evento_nickname in lista])

    sqlcmd_ct_aux = """
            -- MM_XXXXXX table
            CREATE TABLE IF NOT EXISTS IgorMM_{} (
                -- Events id columns
                {},
                P real,
                -- Foreign keys
                {}
            );
            """

    sqlcmd_ct = sqlcmd_ct_aux.format(event_nickname, sqlcmd_table_fields_ct, sqlcmd_foreign_keys_ct)
    return sqlcmd_ct

def sqlcmd_ct_Model_Marginals_DinucMarkov(event_nickname, list_dependencies:list):
    # from the xarray get the list of events something like:
    # lista = [event_nickname]+list_dependencies
    lista = list_dependencies
    str_column_ct = "id_{} integer NOT NULL" #.format(event_nickname)
    str_foreign_key_ct = "FOREIGN KEY (id_{}) REFERENCES IgorER_{} (id)" #.format(event_nickname)

    sqlcmd_table_fields_ct = ",\n".join([str_column_ct.format(evento_nickname) for evento_nickname in lista])
    sqlcmd_foreign_keys_ct = ",\n".join([str_foreign_key_ct.format(evento_nickname, event_nickname) for evento_nickname in lista])

    sqlcmd_ct_aux = """
            -- MM_XXXXXX table
            CREATE TABLE IF NOT EXISTS IgorMM_{} (
                -- Events id columns
                {},
                P real,
                -- Foreign keys
                {}
            );
            """

    sqlcmd_ct = sqlcmd_ct_aux.format(event_nickname, sqlcmd_table_fields_ct, sqlcmd_foreign_keys_ct)
    return sqlcmd_ct

############## IGOR BEST SCENARIOS ##############
sqlcmd_ct['Pgen'] = """
-- IgorPgen table
CREATE TABLE IF NOT EXISTS IgorPgen (
    seq_index integer PRIMARY KEY,
    Pgen_estimate real NOT NULL
);
"""


sqlcmd_ct['BestScenario_template'] = """
-- IgorBestScenarios table
CREATE TABLE IF NOT EXISTS IgorBestScenarios (
    seq_index integer NOT NULL,
    scenario_rank integer NOT NULL,
    scenario_proba_cond_seq real NOT NULL,
    {},
    mismatches text,
    mismatcheslen integer,
    {}
);
"""

def sqlcmd_ct_BestScenarios(nickname_event_type_list:list):
    """
    param nickname_event_type_list: list of tuples (nickname, event_type)
    return sql command to create BestScenarios table.
    """
    # FIXME: ID EVENTS EXCEPT WITH DinucMarkov
    # SELECT * from IgorMP_Event_list; or SELECT nickname, event_type from IgorMP_Event_list;
    # SELECT COUNT(*) FROM IgorMP_Edges WHERE child_event =='d_gene';
    # SELECT COUNT(IgorMP_Edges.parent_event) AS nparents, IgorMP_Event_list.nickname FROM IgorMP_Edges, IgorMP_Event_list WHERE IgorMP_Edges.child_event = IgorMP_Event_list.nickname GROUP BY IgorMP_Event_list.nickname;

    str_column_ct_with_id = "id_{} integer NOT NULL"
    str_column_ct_DinucMarkov = "{} text"
    str_foreign_key_ct = "FOREIGN KEY (id_{}) REFERENCES IgorER_{} (id)"

    str_column_list = list()
    str_foreign_key_list = list()
    for nickname, event_type in nickname_event_type_list:
        if event_type == 'DinucMarkov':
            str_column_list.append(str_column_ct_DinucMarkov.format(nickname))
        else:
            str_column_list.append(str_column_ct_with_id.format(nickname))
            str_foreign_key_list.append(str_foreign_key_ct.format(nickname, nickname))

    str_columns = ",\n".join(str_column_list)
    str_foreign_keys = ",\n".join(str_foreign_key_list)

    sqlcmd_ct_aux = """
    -- Best_scenarios table
    CREATE TABLE IF NOT EXISTS IgorBestScenarios (
        seq_index integer NOT NULL,
        scenario_rank integer NOT NULL,
        scenario_proba_cond_seq real NOT NULL,
        {},
        mismatches text,
        mismatcheslen integer,
        {}
    );
    """

    return sqlcmd_ct_aux.format(str_columns, str_foreign_keys)

# usage : sql_cmd_template_qry_cols.format(tablename='IgorMM_d_3_del')
sqlcmd_template_qry_cols = "SELECT name FROM pragma_table_info('{tablename}');"

# sqlcmd_qry = dict()
# sqlcmd_qry['IgorMarginals'] =

# sqlcmd_ct['ER_XXXXXX'] = """
# -- MP_Event_list table
# CREATE TABLE IF NOT EXISTS MM_XXXXXX (
#     XXXXX_id integer NOT NULL,
#     realization_value text,
#     realization_name text
# );
# """



# # XXX: New tables use these.
# sqlcmd_ct['XXXXXXXXXXXXXXXX'] = """
# YYYYYYYYYYYYYYYYYYYYYYY
# """
#
# sqlcmd_ct['XXXXXXXXXXXXXXXX'] = """
# YYYYYYYYYYYYYYYYYYYYYYY
# """
#
# sqlcmd_ct['XXXXXXXXXXXXXXXX'] = """
# YYYYYYYYYYYYYYYYYYYYYYY
# """
#
# sqlcmd_ct['MM_XXXXXX'] = """
# -- MM_XXXXXX table
# CREATE TABLE IF NOT EXISTS MM_XXXXXX (
#     XXXXX_id integer NOT NULL,
#     P real,
#     FOREIGN KEY (XXXXX_id) REFERENCES ER_XXXXXX (id),
# );
# """

# --------------------------- SQL TABLENAME PATTERNS --------------------------- #

sql_tablename_patterns = ['IgorIndexedSeq',
                          'IgorIndexedCDR3', 'Igor%GeneTemplate', 'Igor%GeneCDR3Anchors', 'Igor%Alignments',
                          'IgorMP_%', 'IgorER_%', 'IgorMM_%',
                          'IgorPgen', 'IgorBestScenarios']

sql_tablename_patterns_dict = dict()
sql_tablename_patterns_read_seqs = ['IgorIndexedSeq']
sql_tablename_patterns_ref_genome = ['Igor%GeneTemplate', 'Igor%GeneCDR3Anchors']
sql_tablename_patterns_align = ['IgorIndexedCDR3', 'Igor%Alignments']
sql_tablename_patterns_model = ['IgorMP_%', 'IgorER_%', 'IgorMM_%']
sql_tablename_patterns_output = ['IgorPgen', 'IgorBestScenarios']
sql_tablename_patterns_pgen = ['IgorPgen']
sql_tablename_patterns_scenarios = ['IgorBestScenarios']
sql_tablename_patterns_dict['read_seqs'] = sql_tablename_patterns_read_seqs
sql_tablename_patterns_dict['ref_genome'] = sql_tablename_patterns_ref_genome
sql_tablename_patterns_dict['align'] = sql_tablename_patterns_align
sql_tablename_patterns_dict['model'] = sql_tablename_patterns_model
sql_tablename_patterns_dict['output'] = sql_tablename_patterns_output
sql_tablename_patterns_dict['pgen'] = sql_tablename_patterns_pgen
sql_tablename_patterns_dict['scenarios'] = sql_tablename_patterns_scenarios

# ----------------------------------------------------------------------------- #





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct  9 16:52:51 2019

@author: alfaceor
"""
#import IgorModel
import pandas as pd
import numpy as np
import sqlite3
from .IgorIO import *

"""
seq_index
scenario_rank
scenario_proba_cond_seq
GeneChoice_V_gene_Undefined_side_prio7_size35
GeneChoice_J_gene_Undefined_side_prio7_size14
GeneChoice_D_gene_Undefined_side_prio6_size2
Deletion_V_gene_Three_prime_prio5_size21
Deletion_D_gene_Five_prime_prio5_size21
Deletion_D_gene_Three_prime_prio5_size21
Deletion_J_gene_Five_prime_prio5_size23
Insertion_VD_genes_Undefined_side_prio4_size31
DinucMarkov_VD_genes_Undefined_side_prio3_size16
Insertion_DJ_gene_Undefined_side_prio2_size31
DinucMarkov_DJ_gene_Undefined_side_prio1_size16
Mismatches
"""




class IgorSqliteDBBestScenariosVDJ:
    """
    Class to create and load table or database with sequences
    """
    def __init__(self, flnIgorSQLBestScenarios):
        #IgorInferenceDB_VDJ.sql
        # FIXME: LOAD DATA FROM PACKAGE ROOT
        self.flnIgorSQLBestScenarios = flnIgorSQLBestScenarios #"IgorDBBestScenariosVDJ.sql"
        self.flnIgorDB  = "" #flnIgorDB
        
        self.conn               = None
    
    def createSqliteDB(self, flnIgorDB):
        """
        Create a SQLite database with the flnIgorDB sql script.
        """
        self.flnIgorDB = flnIgorDB
        self.conn = None

        try:
            self.conn = sqlite3.connect(flnIgorDB)
            qry = open(self.flnIgorSQLBestScenarios, 'r').read()
            cur = self.conn.cursor()
            cur.executescript(qry)
            self.conn.commit()
            cur.close()
            #self.conn.close()
        except sqlite3.Error as e:
            print(e)
    
    
    ###############################################
    ####  IgorBestScenarios Table Methods
    ###############################################
    
    def load_IgorBestScenariosVDJ_FromCSV(self, flnIgorBestScenarios):
        """
        Insert bestScenarios database in database from csv file.
        :param flnIgorBestScenarios:
        :return:
        """
        self.flnIgorBestScenarios = flnIgorBestScenarios        
        
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')
            with open(self.flnIgorBestScenarios) as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    #print(csvline)
                    self.insert_IgorBestScenariosVDJ_FromCSVline(cur, csvline)
            cur.execute('COMMIT')
            #self.conn.commit()
        except sqlite3.Error as e:
            print(e)

    
    def insert_IgorBestScenariosVDJ_FromCSVline(self, cur, csvline):
        """
        Insert IGoR indexed_sequences on Database flnIgorDB
        :param csvline:
        """
        sql = ''' INSERT INTO IgorDBBestScenariosVDJ (
                        seq_index,
                        scenario_rank,
                        scenario_proba_cond_seq,
                        id_v_choice,
                        id_j_choice,
                        id_d_gene,
                        id_v_3_del,
                        id_d_5_del,
                        id_d_3_del,
                        id_j_5_del,
                        id_vd_ins,
                        vd_dinucl,
                        id_dj_ins,
                        dj_dinucl,
                        mismatches,
                        mismatcheslen
                    ) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?) '''
        
        csvline = csvline.replace('\n','')
        csvlist = csvline.split(";")
        for ii, csvelem in enumerate(csvlist):
            if ii in [11, 13, 14]:
                csvlist[ii] = csvlist[ii].replace("(", "[").replace(")", "]")  # lists vd_dinucl, dj_dinucl, mismatches
            else:
                csvlist[ii] = csvlist[ii].replace("(", "").replace(")", "")
        
        if len(csvlist) == 15:
            csvlist.append(str(len(eval(csvlist[14])))) # mismatcheslen
            data = tuple(csvlist)
            try:
                #cur = self.conn.cursor()
                cur.execute(sql, data)
                #self.conn.commit()
            except sqlite3.Error as e:
                print(csvlist)
                print(e)
                pass
        

    def fetch_IgorBestScenariosVDJ_By_seq_index(self, seq_index):
        """
        Fetch seq_index and sequence in Igor database.
        :param seq_index: string to specify the type of gene V, D or J
        :return: 
        """

        #print(gene_name)
        sqlSelect = "SELECT * FROM IgorDBBestScenariosVDJ WHERE seq_index = "+str(seq_index)+";"
        #print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        #if strGene == 'D':
        #    print(sqlSelect)
        record = cur.fetchall()
        return (record)
    
    
    
    
        

#    
#    def __init__(self, flnBestScenarios, model_parms_file=None, model_marginals_file=None):
#        
#        self.seq_index = XXXX
#        self.scenario_rank = XXXX
#        self.scenario_proba_cond_seq = XXXX
#        self.GeneChoice_V_gene_Undefined_side_prio7_size35 = XXXX
#        self.GeneChoice_J_gene_Undefined_side_prio7_size14 = XXXX
#        self.GeneChoice_D_gene_Undefined_side_prio6_size2 = XXXX
#        self.Deletion_V_gene_Three_prime_prio5_size21 = XXXX
#        self.Deletion_D_gene_Five_prime_prio5_size21 = XXXX
#        self.Deletion_D_gene_Three_prime_prio5_size21 = XXXX
#        self.Deletion_J_gene_Five_prime_prio5_size23 = XXXX
#        self.Insertion_VD_genes_Undefined_side_prio4_size31 = XXXX
#        self.DinucMarkov_VD_genes_Undefined_side_prio3_size16 = XXXX
#        self.Insertion_DJ_gene_Undefined_side_prio2_size31 = XXXX
#        self.DinucMarkov_DJ_gene_Undefined_side_prio1_size16 = XXXX
#        self.Mismatches = XXXX
#        
#        
#        self.flnBestScenarios = flnBestScenarios
#        self.mdl = IgorModel.Model(model_parms_file=model_parms_file, model_marginals_file=model_marginals_file)
#        
    def setModelFromFiles(self, model_parms_file=None, model_marginals_file=None):
        self.mdl = IgorModel(model_parms_file=model_parms_file, model_marginals_file=model_marginals_file)
        
    def setModel(self, mdl):
            self.mdl = mdl
    
    def setInputSeqsFile(self, flnInputSeqs):
        self.flnInputSeqs = flnInputSeqs
    
    def setAlignsFile(self, flnAligns):
        self.flnAligns = flnAligns
    
    # TODO: clomplete this if worth it...
    def get_BestScenariosDataFrame(self, flnBestScenarios, flnParms, flnMarginals):
        bestScen = pd.read_csv(flnBestScenarios, sep=';')
        mdl = IgorModel(model_parms_file=flnParms, model_marginals_file=flnMarginals)
        pdBestScen = bestScen.rename(mdl.parms.dictNameNickname, axis=1).set_index('seq_index')
        tmpEvents = ['vd_dinucl', 'dj_dinucl', 'Mismatches']
        tmpEvents02 = ['scenario_rank', 'scenario_proba_cond_seq']
        #strEvent  = 'v_choice'
        for strEvent in pdBestScen.keys():
            if strEvent in tmpEvents:
                pdBestScen[strEvent] = pdBestScen[strEvent].apply(lambda x : x.replace("(", "[").replace(")", "]")).apply(eval)
            elif strEvent in tmpEvents02:
                print(strEvent)
            else:
                #print strEvent
                pdBestScen[strEvent] = pdBestScen[strEvent].apply(eval)

        return pdBestScen

    # Reconstruct the naive sequence from the bestscenarios data.
    def getV_Region(self, seq_id, pdSelected):
        #seq_id=59
        strEv  = 'v_choice'
        id_V   = pdSelected.loc[seq_id][strEv]
        seq_V  = self.mdl.parms.Event_dict[strEv].loc[id_V]['value']#.values    
        name_V = self.mdl.parms.Event_dict[strEv].loc[id_V]['name']
        strEv  = 'v_3_del'
        n_v_3_del = pdSelected.loc[seq_id][strEv]
        #print(name_V, n_v_3_del, seq_V[:-n_v_3_del])
        return (seq_V[:-n_v_3_del])


    def getD_Region(self, seq_id, pdSelected):
        #seq_id=59
        strEv  = 'd_gene'
        id_D   = pdSelected.loc[seq_id][strEv]
        seq_D  = self.mdl.parms.Event_dict[strEv].loc[id_D]['value']#.values    
        name_D = self.mdl.parms.Event_dict[strEv].loc[id_D]['name']
        strEv  = 'd_5_del'
        n_d_5_del = pdSelected.loc[seq_id][strEv]
        strEv  = 'd_3_del'
        n_d_3_del = pdSelected.loc[seq_id][strEv]
        #print(seq_D,len(seq_D))
        #print(seq_id, seq_D,len(seq_D), name_D, n_d_5_del, n_d_3_del, seq_D[n_d_5_del:-n_d_3_del])
        return (seq_D[n_d_5_del:-n_d_3_del])

    def getJ_Region(self, seq_id, pdSelected):
        #seq_id=59
        strEv  = 'j_choice'
        id_J   = pdSelected.loc[seq_id][strEv]
        seq_J  = self.mdl.parms.Event_dict[strEv].loc[id_J]['value']#.values    
        name_J = self.mdl.parms.Event_dict[strEv].loc[id_J]['name']
        strEv  = 'j_5_del'
        n_j_5_del = pdSelected.loc[seq_id][strEv]
        #print(name_J, n_j_5_del, seq_J[n_j_5_del:])
        return (seq_J[n_j_5_del:])

    def getVD_Region(self, seq_id, pdSelected):
        #seq_id=59
        #vd_dinucl
        #vd_ins
        strEv      = 'vd_ins'
        id_VDins   = pdSelected.loc[seq_id][strEv]
        n_VDins  = self.mdl.parms.Event_dict[strEv].loc[id_VDins]['value']#.values    
        #name_VDins = self.mdl.parms.Event_dict[strEv].loc[id_VDins]['name']
        #print(self.mdl.parms.Event_dict[strEv] ) #.loc[id_VDins])
        #print(n_VDins)
        strEv        = 'vd_dinucl'
        id_VD_dinucl = pdSelected.loc[seq_id][strEv]
        seq_VD_dinucl  = self.mdl.parms.Event_dict[strEv].loc[id_VD_dinucl]['value'].values   
        return (''.join(seq_VD_dinucl.tolist()))

    def getDJ_Region(self, seq_id, pdSelected):
        strEv        = 'dj_dinucl'
        id_DJ_dinucl = pdSelected.loc[seq_id][strEv]
        seq_DJ_dinucl  = self.mdl.parms.Event_dict[strEv].loc[id_DJ_dinucl]['value'].values   
        return (''.join(seq_DJ_dinucl.tolist()))
    
    def get_BestScenariosNaiveSeq(self, seq_id, pdBestScen):
        self.getV_Region(seq_id, pdBestScen)
        V_Region  = self.getV_Region(seq_id, pdSelected)
        VD_Region = self.getVD_Region(seq_id, pdSelected)
        D_Region  = self.getD_Region(seq_id, pdSelected)
        DJ_Region = self.getDJ_Region(seq_id, pdSelected)
        J_Region  = self.getJ_Region(seq_id, pdSelected)
        return V_Region+VD_Region+D_Region+DJ_Region+J_Region
        




import collections
def from_fasta_to_dataframe(fln_fasta):
    # Fasta to dataframe
    from Bio import SeqIO
    import pandas as pd
    genes_name_list = list()
    genes_value_list = list()
    for gene_record in SeqIO.parse(fln_fasta, "fasta"):
        genes_name_list.append(gene_record.description)
        genes_value_list.append(str(gene_record.seq))
    df_genes = pd.DataFrame.from_dict({'name': genes_name_list, 'value': genes_value_list})
    df_genes.index.name = 'id'
    return df_genes



# // A, C, G, T, R, Y, K, M, S, W, B, D, H, V, N
heavy_pen_nuc44_vect = [
5, -14, -14, -14, -14, 2, -14, 2, 2, -14, -14, 1, 1, 1, 0,
-14, 5, -14, -14, -14, 2, 2, -14, -14, 2, 1, -14, 1, 1, 0,
-14, -14, 5, -14, 2, -14, 2, -14, 2, -14, 1, 1, -14, 1, 0,
-14, -14, -14, 5, 2, -14, -14, 2, -14, 2, 1, 1, 1, -14, 0,
-14, -14, 2, 2, 1.5, -14, -12, -12, -12, -12, 1, 1, -13, -13, 0,
2, 2, -14, -14, -14, 1.5, -12, -12, -12, -12, -13, -13, 1, 1, 0,
-14, 2, 2, -14, -12, -12, 1.5, -14, -12, -12, 1, -13, -13, 1, 0,
2, -14, -14, 2, -12, -12, -14, 1.5, -12, -12, -13, 1, 1, -13, 0,
2, -14, 2, -14, -12, -12, -12, -12, 1.5, -14, -13, 1, -13, 1, 0,
-14, 2, -14, 2, -12, -12, -12, -12, -14, 1.5, 1, -13, 1, -13, 0,
-14, 1, 1, 1, 1, -13, 1, -13, -13, 1, 0.5, -12, -12, -12, 0,
1, -14, 1, 1, 1, -13, -13, 1, 1, -13, -12, 0.5, -12, -12, 0,
1, 1, -14, 1, -13, 1, -13, 1, -13, 1, -12, -12, 0.5, -12, 0,
1, 1, 1, -14, -13, 1, 1, -13, 1, -13, -12, -12, -12, 0.5, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

import xarray as xr
import numpy as np
list_nt_lbl = ['A', 'C', 'G', 'T', 'R', 'Y', 'K', 'M', 'S', 'W', 'B', 'D', 'H', 'V', 'N']
da_heavy_pen_nuc44_vect = xr.DataArray(np.array(heavy_pen_nuc44_vect).reshape(15, 15), \
                               dims=('x', 'y'))
# print(len(list_nt_lbl))
strDim = 'x'
da_heavy_pen_nuc44_vect[strDim] = range(len(list_nt_lbl))
strCoord = 'lbl__' + strDim
da_heavy_pen_nuc44_vect[strCoord] = (strDim, list_nt_lbl)

strDim = 'y'
da_heavy_pen_nuc44_vect[strDim] = range(len(list_nt_lbl))
strCoord = 'lbl__' + strDim
da_heavy_pen_nuc44_vect[strCoord] = (strDim, list_nt_lbl)



#####################################################
########## PLOTTING SEQUENCES  #################
#####################################################
# Functions copied from :
# Reference: https://dmnfarrell.github.io/bioinformatics/bokeh-sequence-aligner

# FIXME: PLEASE MAKE IT MORE MATPLOTLIB-ISH
try:
    from Bio import AlignIO, SeqIO
    import numpy as np
    from bokeh.plotting import figure, output_file, save, show
    from bokeh.models import ColumnDataSource, Plot, Grid, Range1d
    from bokeh.models.glyphs import Text, Rect
    from bokeh.layouts import gridplot

    def get_colors(seqs):
        """make colors for bases in sequence"""
        text = [i for s in list(seqs) for i in s]
        clrs = {'A': 'red', 'T': 'green', 'G': 'orange', 'C': 'blue', '-': 'white',
                'a': 'red', 't': 'green', 'g': 'orange', 'c': 'blue'}
        colors = [clrs[i] for i in text]
        return colors


    def view_alignment(aln, fontsize="9pt", plot_width=800):
        """Bokeh sequence alignment view"""
        # make sequence and id lists from the aln object
        seqs = [rec.seq for rec in (aln)]
        ids = [rec.id for rec in aln]
        text = [i for s in list(seqs) for i in s]

        colors = get_colors(seqs)
        N = len(seqs[0])
        S = len(seqs)
        width = .4
        x = np.arange(1, N + 1)
        y = np.arange(0, S, 1)
        # creates a 2D grid of coords from the 1D arrays
        xx, yy = np.meshgrid(x, y)
        # flattens the arrays
        gx = xx.ravel()
        gy = yy.flatten()
        # use recty for rect coords with an offset
        recty = gy + .5
        h = 1 / S
        # now we can create the ColumnDataSource with all the arrays
        source = ColumnDataSource(dict(x=gx, y=gy, recty=recty, text=text, colors=colors))
        plot_height = len(seqs) * 15 + 50
        x_range = Range1d(0, N + 1, bounds='auto')
        if N > 100:
            viewlen = 100
        else:
            viewlen = N
        # view_range is for the close up view
        view_range = (0, viewlen)
        tools = "xpan, xwheel_zoom, reset, save"

        # entire sequence view (no text, with zoom)
        p = figure(title=None, plot_width=plot_width, plot_height=50,
                   x_range=x_range, y_range=(0, S), tools=tools,
                   min_border=0, toolbar_location='below')
        rects = Rect(x="x", y="recty", width=1, height=1, fill_color="colors",
                     line_color=None, fill_alpha=0.6)
        p.add_glyph(source, rects)
        p.yaxis.visible = False
        p.grid.visible = False

        # sequence text view with ability to scroll along x axis
        p1 = figure(title=None, plot_width=plot_width, plot_height=plot_height,
                    x_range=view_range, y_range=ids, tools="xpan,reset",
                    min_border=0, toolbar_location='below')  # , lod_factor=1)
        glyph = Text(x="x", y="y", text="text", text_align='center', text_color="black",
                     text_font="monospace", text_font_size=fontsize)
        rects = Rect(x="x", y="recty", width=1, height=1, fill_color="colors",
                     line_color=None, fill_alpha=0.4)
        p1.add_glyph(source, glyph)
        p1.add_glyph(source, rects)

        p1.grid.visible = False
        p1.xaxis.major_label_text_font_style = "bold"
        p1.yaxis.minor_tick_line_width = 0
        p1.yaxis.major_tick_line_width = 0

        p = gridplot([[p], [p1]], toolbar_location='below')
        show(p)
        return p


except ImportError as error:
    # Output expected ImportErrors.
    # print(error.__class__.__name__ + ": " + error.message)
    # FIXME: no error printing for autocomplete
    # print(error)
    pass
except Exception as exception:
    # Output unexpected Exceptions.
    print(exception, False)
    print(exception.__class__.__name__ + ": " + exception.message)


def run_igor_datadir():
    import subprocess

    cmd = "which igor"
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    line = p.stdout.readline()
    igor_exec_path = line.decode("utf-8").replace('\n', '')

    cmd = igor_exec_path + " -getdatadir"

    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout = []
    while True:
        line = p.stdout.readline()
        line = line.decode("utf-8")
        stdout.append(line)
        # print (line, end='')
        if line == '' and p.poll() != None:
            break
    return (''.join(stdout)).replace('\n','')


class GeneSegment:
    def __int__(self, gene_type=None):
        self.gene_type = gene_type
        self.palindrome_5_end = None
        self.gene_ini = None
        self.gene_end = None
        self.gene_cut = None
        self.palindrome_3_end = None
        self.gene_segment = None

class InsertSegment:
    def __int__(self, gene_type=None):
        self.gene_type = gene_type
        self.palindrome_5_end = None
        self.gene_ini = None
        self.gene_end = None
        self.gene_cut = None
        self.palindrome_3_end = None
        self.gene_segment = None

def get_gene_segment(str_gene_template, int_gene_5_del=None, int_gene_3_del=None):
    if int_gene_5_del is None:
        int_gene_5_del = 0
    if int_gene_3_del is None:
        int_gene_3_del = 0

    # print(str_gene_template, int_gene_5_del, int_gene_3_del)

    int_ini = 0
    int_end = len(str_gene_template)
    str_gene_3_palindrome = ""
    str_gene_5_palindrome = ""

    if int_gene_5_del < 0:
        int_ini = 0
        str_gene_5_palindrome = dna_complementary( (str_gene_template[:-int_gene_5_del])[::-1] )
    else:
        int_ini = int_gene_5_del
        str_gene_5_palindrome = ""

    if int_gene_3_del < 0:
        int_end = len(str_gene_template)
        str_gene_3_palindrome = dna_complementary( (str_gene_template[int_gene_3_del:])[::-1] )
    else:
        int_end = len(str_gene_template) - int_gene_3_del
        str_gene_3_palindrome = ""

    segment_dict = collections.OrderedDict()
    segment_dict['palindrome_5_end'] = str_gene_5_palindrome
    segment_dict['gene_ini'] = int_ini
    segment_dict['gene_end'] = int_end
    segment_dict['gene_cut'] = str_gene_template[int_ini:int_end]
    segment_dict['palindrome_3_end'] = str_gene_3_palindrome
    segment_dict['gene_segment'] = str_gene_5_palindrome + str_gene_template[int_ini:int_end] + str_gene_3_palindrome
    return segment_dict
    # str_gene_segment = str_gene_5_palindrome + str_gene_template[int_ini:int_end] + str_gene_3_palindrome
    # return str_gene_segment


def dna_complementary(str_seq):
    from Bio.Seq import Seq
    return str(Seq(str_seq).complement())



#      Author: Quentin Marcou & Carlos Olivares
#
#  This source code is distributed as part of the IGoR software.
#  IGoR (Inference and Generation of Repertoires) is a versatile software to
#  analyze and model immune receptors generation, selection, mutation and all
#  other processes.
#   Copyright (C) 2017  Quentin Marcou
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

###   from . import IgorAligns
###   from . import aligns
###   from . import counters
###   from . import models
###   from . import utils
###   
###   __all__ = ["aligns","counters","models","utils"]


###   from . import IgorAlignment_data
###   from . import IgorModel
###   from . import IgorBestScenarios
###   from . import IgorIndexedSequencesDB
###   from . import IgorSqliteDBBestScenarios
###   from . import IgorSqliteDB


###   import IgorAlignment_data
###   import IgorBestScenarios
###   import IgorIndexedSequencesDB
###   import IgorModel
###   import IgorSqliteDBBestScenarios
###   import IgorSqliteDB
###   __all__ = ["IgorAlignment_data","IgorBestScenarios","IgorIndexedSequencesDB","IgorModel", "IgorSqliteDBBestScenarios", "IgorSqliteDB"]



class RcParams(dict):
    #validate = { key : converter
    #            for key, (default, converter) in defaultParams.items()}
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __setitem__(self, key, val):
        #try:
        #    cval = self.validate[key](val)
        #except ValueError as ve:
        #    raise ValueError(f"Key {key} : {ve}") from None
        dict.__setitem__(self, key, val)
    
    def __getitem__(self, key):
        return dict.__getitem__(self, key)

defaultParams = {
    # default system paths
    'paths.igor_prefix' : None, #"/home/alfaceor/.local",
    'paths.igor_exec' : None, #"/home/alfaceor/.local/bin/igor",
    'paths.igor_data' : None, #"",
    'paths.igor_models': None, #"/home/alfaceor/.local/share/igor/models",
    'paths.igor_src' : None #""
}

rcParams = RcParams(defaultParams)

import pkg_resources

# __name__ in case you're within the package
# - otherwise it would be 'lidtk' in this example as it is the package name
sql_path = 'IgorDB.sql'  #
sql_filepath = pkg_resources.resource_filename(__name__, sql_path)

#from .IgorModel import *
from .IgorIO import *
from .IgorSqliteDB import *
# from .IgorSqliteDBBestScenarios import *
# from .IgorBestScenarios import *
from . import imgt

from .AIRR import *

#__all__ = ["IgorModel"]





#!/usr/bin/python3

# TODO: Script to change the indexed_seqs and the alignment files into a 
# a database to check and access the data
# So also I need to add a option to IGoR to get the data used for the alignment
# and inference in order to integrated with this new pygor3
# - check how to made the relational tables. How to set FOREIGN KEYS in sqlite

import sqlite3

class IgorIndexedSequencesDB:
    """
    Class to create and load table or database with sequences
    """
    def __init__(self):
        flnIndexedSeqDB=""
    
    """
    :return IgorIndexedSequencesDB loaded with a table on a specified database.
    """
    @classmethod
    def loadDataBase(cls, flnDB):
        print("DataBase")
        
    
#-- IgorIndexedSeq table
create_table_sql = '''
CREATE TABLE IF NOT EXISTS IgorIndexedSeq (
    seq_index integer PRIMARY KEY,
    sequence text NOT NULL
);
'''

#-- IgorVGeneTemplate table
'''
CREATE TABLE IF NOT EXISTS IgorVGeneTemplate (
    vgene_id integer PRIMARY KEY,
    gene_name text NOT NULL,
    sequence text NOT NULL,
);
'''


#-- IgorVAlignments table
'''
CREATE TABLE IF NOT EXISTS IgorVAlignments (
    id integer PRIMARY KEY,
    seq_index integer,
    vgene_id integer,
    score real,
    offset integer,
    insertions text NOT NULL,
    deletions  text NOT NULL,
    mismatches text NOT NULL,
    length integer,
    offset_5_p_align integer,
    offset_3_p_align integer,
    FOREIGN KEY (project_id) REFERENCES projects (id)
);
'''

def create_connection(db_file):
    """ create a database connection to the SQLite database
        specified by db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except sqlite3.Error as e:
        print(e)
 
    return conn


def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except sqlite3.Error as e:
        print(e)

def insert_IgorIndexedSeq_FromCSVline(conn, csvline):
    """
    Create a new task
    :param conn:
    :param csvline:
    :return:
    """
 
    sql = ''' INSERT INTO IgorIndexedSeq(seq_index,sequence)
              VALUES(?,?) '''
    
    data = tuple(csvline.split(";"))
    try:
        cur = conn.cursor()
        print(data)
        cur.execute(sql, data)
        conn.commit()
    except sqlite3.Error as e:
        print(e)
        pass

#cursor.execute('''INSERT INTO Simulation(Directory, strF2B, strB2F, strB2B, strHrad, strBrad) VALUES(?, ?, ?, ?, ?, ?)''',



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct  3 11:02:28 2019

@author: alfaceor
"""

import sqlite3
import csv
from Bio import SeqIO
#import IgorAlignment_data
from .IgorSQL import *
import numpy as np
import pandas as pd

###################
# TODO: Generalize to other kinds of database like postgres
###################

class IgorSqliteDB:
    """
    Class to create and load table or database with sequences
    """
    def __init__(self): #=None):
        # if flnIgorSQL is None:
        #     self.flnIgorSQL = "IgorDB.sql" # FIXME : VERY BAD WRAPP WITH INIT
        # else:
        #     self.flnIgorSQL = flnIgorSQL
        self.flnIgorSQL = "" # FIXME: this shouldn't be need it!

        self.flnIgorDB  = "" #flnIgorDB
        
        self.flnIgorIndexedSeq = ""
        self.flnIgorIndexedCDR3 = ""
        
        self.flnVGeneTemplate   = ""
        self.flnDGeneTemplate   = ""
        self.flnJGeneTemplate   = ""
        
        self.flnVAlignments     = ""
        self.flnDAlignments     = ""
        self.flnJAlignments     = ""
        
        self.flnVAnchors        = ""
        self.flnJAnchors        = ""

        self.flnIgorModel_Parms = ""
        self.flnIgorModel_Marginals = "" #

        self.flnIgorBestScenarios = ""
        self.flnIgorPgen = ""
        
        self.conn               = None

        self.sql_BestScenarios_cols_list = None
        self.sqlcmd_ins_bs = None
        self.sql_IgorBestScenarios_cols_name_type_list = None
        self.mdl = None
    
    #@classmethod
    def createSqliteDB(self, flnIgorDB):
        # FIXME: THIS METHOD SHOULD BE SOMETHING LIKE EXECUTE SQL SCRIPT and in particular
        """
        Create a SQLite database with the flnIgorDB sql script.
        """
        self.flnIgorDB = flnIgorDB
        self.conn = None
        try:
            self.conn = sqlite3.connect(flnIgorDB)
            qry = open(self.flnIgorSQL, 'r').read()
            cur = self.conn.cursor()
            cur.executescript(qry)
            self.conn.commit()
            cur.close()
            #self.conn.close()
        except sqlite3.Error as e:
            print(e)

    @classmethod
    def create_db(cls, flnIgorDB):
        """
        Connect (or create if not exits) with filename
        """
        cls = IgorSqliteDB()
        cls.flnIgorDB = flnIgorDB
        cls.connect_db()
        return cls

    def connect_db(self, flnIgorDB=None):
        """
        Connect (or create if not exits) to database
        """
        if flnIgorDB is not None:
            self.flnIgorDB = flnIgorDB
        self.conn = sqlite3.connect(self.flnIgorDB)

    def close_db(self):
        self.conn.close()

    def executescript(self, cur, str_query):
        cur.executescript(str_query)

    def execute_query(self, str_query):
        """
        Execute sql script in the SQLite database .
        """
        try:
            self.conn = sqlite3.connect(self.flnIgorDB)
            cur = self.conn.cursor()
            cur.executescript(str_query)
            self.conn.commit()
            cur.close()
            # self.conn.close()
        except sqlite3.Error as e:
            print('str_query : ', str_query)
            print("sqlite3.ERROR : ", e)

    def execute_select_query(self, str_query):
        """
        Execute sql script in the SQLite database .
        """
        try:
            self.conn = sqlite3.connect(self.flnIgorDB)
            cur = self.conn.cursor()
            cur.execute(str_query)
            # self.conn.commit()
            record = cur.fetchall()
            # print(record)
            # cur.close()
            return record
            # self.conn.close()
        except sqlite3.Error as e:
            print("sqlite3.ERROR : ", e)

    def execute_select_query_fetchone(self, str_query):
        """
        Execute sql script in the SQLite database .
        """
        try:
            self.conn = sqlite3.connect(self.flnIgorDB)
            cur = self.conn.cursor()
            cur.execute(str_query)
            #self.conn.commit()
            record = cur.fetchone()
            # print(record)
            # cur.close()
            return record
            #self.conn.close()
        except sqlite3.Error as e:
            print("sqlite3.ERROR : ", e)

    def createSqliteDB_tmp(self):
        # TODO: create database in base of IgorSQL scripts
        self.conn = None
        try:
            self.conn = sqlite3.connect(self.flnIgorDB)
            qry = sqlcmd_ct['indexed_sequences']
            cur = self.conn.cursor()
            cur.executescript(qry)
            self.conn.commit()
            cur.close()
            # self.conn.close()
        except sqlite3.Error as e:
            print(e)

    # TODO: load database by receiving a list of identifiers defined in IgorDictionaries
    def load_db(self, **kwargs):
        """
        Return a parameter
        """
        for key in kwargs.keys():
            # Create a database.
            print( sqlcmd_ct[key] )
            #print(key)
            self.insert_IgorIndexedSeq_FromCSVline()

            # cur = self.conn.cursor()
            # try:
            #     cur.execute('BEGIN TRANSACTION')
            #     with open(flnIgorIndexedSeq) as fp:
            #         csvline = fp.readline()
            #         while csvline:
            #             csvline = fp.readline()
            #             # print(csvline)
            #             self.insert_IgorIndexedSeq_FromCSVline(cur, csvline)
            #     cur.execute('COMMIT')
            #     # self.conn.commit()
            # except sqlite3.Error as e:
            #     print(e)

    def load_VDJ_Database(self, flnIgorIndexedSeq, flnVGeneTemplate, flnDGeneTemplate, flnJGeneTemplate, flnVAlignments, flnDAlignments, flnJAlignments):
        
        self.load_IgorIndexedSeq_FromCSV(flnIgorIndexedSeq )
        
        self.load_IgorGeneTemplate_FromFASTA("V",flnVGeneTemplate)
        self.load_IgorGeneTemplate_FromFASTA("D",flnDGeneTemplate)
        self.load_IgorGeneTemplate_FromFASTA("J",flnJGeneTemplate)
        
        self.load_IgorAlignments_FromCSV("V", flnVAlignments)
        self.load_IgorAlignments_FromCSV("D", flnDAlignments)
        self.load_IgorAlignments_FromCSV("J", flnJAlignments)
    
    ###############################################
    ####  IgorIndexedSeq Table Methods
    ###############################################

    def load_IgorIndexedSeq_FromCSV(self, flnIgorIndexedSeq):
        """
        Insert indexed sequence in database from csv igor indexed_seqs file.
        :param conn:
        :param csvline:
        :return:
        """
        self.execute_query(sqlcmd_ct['indexed_sequences'])
        self.flnIgorIndexedSeq = flnIgorIndexedSeq
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')
            with open(flnIgorIndexedSeq) as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    # print(csvline)
                    self.insert_IgorIndexedSeq_FromCSVline(cur, csvline)
            cur.execute('COMMIT')
            # self.conn.commit()
        except sqlite3.Error as e:
            print(e)

    def insert_IgorIndexedSeq_FromCSVline(self, cur, csvline):
        """
        Insert IGoR indexed_sequences on Database flnIgorDB
        :param csvline:
        """
        sql = ''' INSERT INTO IgorIndexedSeq(seq_index,sequence)
                  VALUES(?,?) '''
        
        csvline = csvline.replace('\n','')
        data = tuple(csvline.split(";"))
        if len(data) == 2:
            try:
                #cur = self.conn.cursor()
                cur.execute(sql, data)
                #self.conn.commit()
            except sqlite3.Error as e:
                print(data)
                print(e)
                pass

    def fetch_IgorIndexedSeq_records(self):
        """
        Fetch seq_index and sequence from Igor database.
        :param seq_index: string to specify the type of gene V, D or J
        :return:
        """

        #print(gene_name)
        sqlSelect = "SELECT seq_index, sequence FROM IgorIndexedSeq;"
        #print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        #if strGene == 'D':
        #    print(sqlSelect)
        records = cur.fetchall()
        return records

    def fetch_IgorIndexedSeq_By_seq_index(self, seq_index):
        """
        Fetch seq_index and sequence in Igor database.
        :param seq_index: string to specify the type of gene V, D or J
        :return: 
        """

        #print(gene_name)
        sqlSelect = "SELECT seq_index, sequence FROM IgorIndexedSeq WHERE seq_index = "+str(seq_index)+";"
        #print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        #if strGene == 'D':
        #    print(sqlSelect)
        record = cur.fetchone()
        return (record)

    def get_IgorIndexedSeq_By_seq_index(self, seq_index):
        record = self.fetch_IgorIndexedSeq_By_seq_index(seq_index)
        from .IgorIO import IgorIndexedSequence
        return IgorIndexedSequence.load_FromSQLRecord(record)

    def fetch_IgorIndexedSeq_By_seq_indexList(self, seq_indexList):
        """
        Fetch seq_index and sequence in Igor database.
        :param seq_index: string to specify the type of gene V, D or J
        :return: 
        """

        #print(gene_name)
        strSeq_indexList = str( tuple( sorted( set( seq_indexList ) ) )  )

        sqlSelect = "SELECT seq_index, sequence FROM IgorIndexedSeq WHERE seq_index IN "+strSeq_indexList+";"
        #print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        #if strGene == 'D':
        #    print(sqlSelect)
        record = cur.fetchall()
        return (record)

    def fetch_IgorIndexedSeq_indexes(self):
        seq_indexes_list = self.execute_select_query("SELECT seq_index FROM IgorIndexedSeq;")
        seq_indexes_list = list( map(lambda x: x[0], seq_indexes_list) )
        return seq_indexes_list

    def write_IgorIndexedSeq_to_CSV(self, flnIgorIndexedSeq, sep=";"):
        print("Saving indexed_sequences to file: ", flnIgorIndexedSeq)
        # print(gene_name)
        sqlSelect = "SELECT seq_index, sequence FROM IgorIndexedSeq;"
        # print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        # if strGene == 'D':
        #    print(sqlSelect)
        records = cur.fetchall()
        with open(flnIgorIndexedSeq, "w") as ofile:
            ofile.write("seq_index" + sep + "sequence" + "\n")
            for record in records:
                ofile.write(str(record[0])+sep+str(record[1])+"\n")

    def delete_IgorIndexedSeq_Tables(self):
        """
        Method to delete IgorIndexedSeq table.
        """
        tables_list = self.get_list_of_tables_with_name('IgorIndexedSeq')
        for tablename in tables_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)

    ###############################################
    ####  IgorXGeneTemplate Tables Methods
    ###############################################

    def load_IgorGeneTemplate_FromFASTA(self, strGene, flnGeneTemplate):
        """
        Insert D Gene templates in database from fasta files used by IGoR.
        :param flnIgorGeneTemplate: Fasta file
        """
        # TODO: ADD A PANDAS OBJECT FOR RAPID ACCESS TO GENOMIC TEMPLATES.
        filename = {'V': self.flnVGeneTemplate, 'D' : self.flnDGeneTemplate, 'J': self.flnJGeneTemplate }
        filename[strGene] = flnGeneTemplate
        print(strGene, filename[strGene])
        # Create table if don't exits
        self.execute_query(sqlcmd_ct['genomic'+strGene + 's'])
        with open(filename[strGene], "r") as handle:
            for gene_id, bioRecord in enumerate(SeqIO.parse(handle, "fasta") ):
                self.insert_IgorGeneTemplate_FromBioRecord(strGene, gene_id, bioRecord)

    def insert_IgorGeneTemplate_FromBioRecord(self, strGene, gene_id, bioRecord):
        """
        Insert IGoR Gene template in Database flnIgorDB
        :param strGene: string to specify the type of gene V, D or J
        :param gene_id: id to identify the gene template
        :param bioRecord: Biopython record of the inserted sequence
        """
        sql = "INSERT INTO Igor"+strGene+"GeneTemplate("+strGene.lower()+"gene_id,gene_name,sequence) VALUES(?,?,?) "

        data = tuple([gene_id, str(bioRecord.description).strip(), str(bioRecord.seq)])
        try:
            cur = self.conn.cursor()
            cur.execute(sql, data)
            self.conn.commit()
        except sqlite3.Error as e:
            print(e)
            pass

    def fetch_IgorGeneTemplate_By_gene_name(self, strGene, gene_name):
        """
        Fetch Gene templates in database from fasta files used by IGoR.
        :param strGene: string to specify the type of gene V, D or J
        :param flnIgorGeneTemplate: Fasta file
        """
        sqlSelect = "SELECT * FROM Igor"+strGene.upper()+"GeneTemplate WHERE gene_name =\""+gene_name+"\";"
        #print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        #if strGene == 'D':
        #    print(sqlSelect)
        record = cur.fetchone()
        return (record)        
    
    def fetch_IgorGeneTemplate_By_gene_id(self, strGene, gene_id):
        """
        Fetch Gene templates in database from fasta files used by IGoR.
        :param strGene: string to specify the type of gene V, D or J
        :param gene_id: 
        """
        sqlSelect = "SELECT * FROM Igor"+strGene.upper()+"GeneTemplate WHERE "+strGene.lower()+"gene_id ="+str(gene_id)+";"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        record = cur.fetchone()
        return (record)

    def write_IgorGeneTemplate_to_fasta(self, strGene, flnGeneTemplate, sep=";"):
        print("Saving gene templates to file: ", flnGeneTemplate)
        sqlSelect = "SELECT * FROM Igor"+strGene.upper()+"GeneTemplate;"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        # str_file_header = "seq_index" + sep + "sequence" + "\n"
        with open(flnGeneTemplate, "w") as ofile:
            # ofile.write(str_file_header)
            for record in records:
                ofile.write(">" + str(record[1]) + "\n")
                ofile.write(str(record[2]) + "\n")

    def delete_IgorGeneTemplate_Tables(self):
        """
        Method to delete IgorIndexedSeq table.
        """
        tables_list = self.get_list_of_tables_with_name('Igor%GeneTemplate')
        print("Tables to delete : ", tables_list)
        for tablename in tables_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)


    def load_IgorGeneAnchors_FromCSV(self, strGene, flnGeneAnchors):
        # import pdb; pdb.set_trace()

        self.execute_query(sqlcmd_ct['gene'+strGene+'CDR3Anchors'])
        # FIXME: FILENAMES WERE NOT STORED
        filename = {'V': self.flnVAnchors, 'J': self.flnJAnchors}
        filename[strGene] = flnGeneAnchors

        print("Loading Gene Anchors from ", flnGeneAnchors)
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')
            with open(filename[strGene], "r") as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    self.insert_IgorGeneAnchors_FromCSVline(strGene, cur, csvline)

            cur.execute('COMMIT')
            # self.conn.commit()
        except sqlite3.Error as e:
            print("sqlite3.ERROR : ", e)

    def insert_IgorGeneAnchors_FromCSVline(self, strGene, cur, csvline):
        """
        Insert IGoR indexed_CDR3_sequences in Database flnIgorDB
        :param csvline:
        """
        sql = "INSERT INTO Igor"+strGene+\
              "GeneCDR3Anchors("+strGene.lower()+"gene_id, anchor_index, function) VALUES(?,?,?) "

        csvline = csvline.replace('\n', '')
        data = csvline.split(";")

        try:
            data[0] = self.fetch_IgorGeneTemplate_By_gene_name(strGene, data[0])[0]
        except Exception as e:
            print("data : ", data)
            print("ERROR : ", e)
            pass

        try:
            if len(data) == 2:
                # NO FUNCTION SPECIFIED
                data = data + [None]
                cur.execute(sql, data)
            elif len(data) == 3:
                cur.execute(sql, data)
            else:
                print(data)

        except sqlite3.Error as e:
            print(len(data), data)
            print("sqlite3.ERROR : ", e)
            pass

    # TODO: fetch records from database
    def fetch_IgorGeneAnchors_By_Gene(self, strGene):
        sqlSelect = "SELECT * FROM Igor" + strGene.upper() + "GeneCDR3Anchors"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        return (records)

    def write_IgorGeneAnchors_to_CSV(self, strGene, flnGeneAnchors, sep=';'):
        print("Saving IGoR's "+strGene+" gene anchors ")
        sqlcmd_select_template = """
                    SELECT gene.gene_name,
                            anch.anchor_index, anch.function 
                    FROM Igor{upper}GeneTemplate gene 
                    LEFT JOIN Igor{upper}GeneCDR3Anchors anch 
                    ON gene.{lower}gene_id = anch.{lower}gene_id;
                    """

        sqlSelect = sqlcmd_select_template.format(upper=strGene.upper(), lower=strGene.lower())

        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        str_file_header = "gene;anchor_index;function\n"
        with open(flnGeneAnchors, "w") as ofile:
            ofile.write(str_file_header)
            for record in records:
                ofile.write(str(record[0]) + sep)
                ofile.write(str(record[1]) + sep)
                ofile.write(str(record[2]) + '\n')

    def delete_IgorGeneAnchors_Tables(self):
        """
        Method to delete IgorIndexedSeq table.
        """
        tables_list = self.get_list_of_tables_with_name('Igor%GeneCDR3Anchors')
        print("Tables to delete : ", tables_list)
        for tablename in tables_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)


    def fetch_IgorGenomicData_By_Gene(self, strGene):
        if strGene == 'D':
            sqlcmd_select_template = """
                                SELECT gene.{lower}gene_id, gene.gene_name, gene.sequence 
                                FROM Igor{upper}GeneTemplate gene;
                                """
        else:
            sqlcmd_select_template = """
                    SELECT gene.{lower}gene_id, gene.gene_name, gene.sequence, 
                            anch.anchor_index, anch.function 
                    FROM Igor{upper}GeneTemplate gene 
                    LEFT JOIN Igor{upper}GeneCDR3Anchors anch 
                    ON gene.{lower}gene_id = anch.{lower}gene_id;
                    """
        sqlSelect = sqlcmd_select_template.format(upper=strGene.upper(), lower=strGene.lower())

        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        return (records)

    def get_IgorGenomicDataFrame_dict(self):
        """return dataframes genomic_data """
        import pandas as pd
        genomic_data_dict = dict()
        V_records = self.fetch_IgorGenomicData_By_Gene("V")
        J_records = self.fetch_IgorGenomicData_By_Gene("J")

        columnas = ['id', 'gene_name', 'sequence', 'anchor_index', 'function']
        df_V = pd.DataFrame.from_records(V_records, columns=columnas)
        genomic_data_dict["V"] = df_V
        df_J = pd.DataFrame.from_records(J_records, columns=columnas)
        genomic_data_dict["J"] = df_J

        try:
            D_records = self.fetch_IgorGenomicData_By_Gene("D")
            columnas = ['id', 'gene_name', 'sequence']
            df_D = pd.DataFrame.from_records(D_records, columns=columnas)
            genomic_data_dict["D"] = df_D
        except Exception as e:
            print("No D genes were found in database ", self.flnIgorDB)
            print("WARNING: ", e)
            pass


        return genomic_data_dict


    ###############################################
    ####  IgorXAlignments Tables Methods
    ###############################################
    
    def load_IgorAlignments_FromCSV(self, strGene, flnAlignments):
        """
        Insert Gene templates in database from fasta files used by IGoR.
        :param strGene: string to specify the type of gene V, D or J
        :param flnIgorGeneTemplate: Fasta file
        """

        # Load database from file
        strGene = strGene.upper()
        filename = {'V': self.flnVAlignments, 'D' : self.flnDAlignments, 'J': self.flnJAlignments}
        filename[strGene] = flnAlignments
        # Create table if don't exits
        self.execute_query(sqlcmd_ct[strGene + '_alignments'])
        try:
            cur = self.conn.cursor()
            cur.execute('BEGIN TRANSACTION')
            with open(filename[strGene]) as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    #print(csvline)
                    self.insert_IgorAlignments_FromCSVline(cur, strGene, csvline)
            
            cur.execute('COMMIT')
        except sqlite3.Error as e:
            print(e)

    def insert_IgorAlignments_FromCSVline(self, cur, strGene, csvline):
        """
        Insert IGoR Alignments on Database flnIgorDB
        :param strGene: string to specify the type of gene V, D or J
        :param csvline:
        """
        sqlBase = '''INSERT INTO Igor{}Alignments(
                seq_index, 
                {}gene_id, 
                score,
                offset,
                insertions,
                deletions,
                mismatches,
                length,
                offset_5_p_align,
                offset_3_p_align) 
                VALUES(?,?,?,?,?,?,?,?,?,?)
                '''
        
        strGene =  strGene.upper()
        sql = sqlBase.format(strGene.upper(), strGene.lower())
        # search on IgorGeneTemplate the corresponding Gene id 
        csvline = csvline.replace("{","[").replace("}","]").replace('\n','')
        csvlist = csvline.split(";")
        if len(csvlist) == 10 :
            gene_name = csvlist[1]
            gene_name = gene_name.strip()
            gene_id = self.fetch_IgorGeneTemplate_By_gene_name(strGene, gene_name)[0]
            #print(gene_id, gene_name)
            csvlist[1] = str(gene_id)
            data = tuple(csvlist)
            #print(data)
            try:
                #cur = self.conn.cursor()
                cur.execute(sql, data)
                #self.conn.commit()
            except sqlite3.Error as e:
                print(e)
                pass
        else:
            print(csvlist)

    def fetch_IgorAlignments_By_seq_index(self, strGene, seq_index, limit=None):
        """
        Fetch IgorAlignments from database by seq_index.
        :param strGene: string to specify the type of gene V, D or J
        :param seq_index: IgorIndexedSequences index
        """
        sqlSelect = "SELECT * FROM Igor"+strGene.upper()+"Alignments WHERE seq_index=="+str(seq_index)+" ORDER BY score DESC"
        if limit is not None:
            sqlSelect = sqlSelect + " LIMIT " + str(limit)
        sqlSelect = sqlSelect +";"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        record = cur.fetchall()
        return (record)

    def fetch_IgorAlignments_By_seq_index_and_gene_name(self, strGene, seq_index, gene_name, limit=None):
        sqlSelect = "SELECT * FROM Igor" + strGene.upper() + "Alignments WHERE seq_index==" + str(
            seq_index) + " ORDER BY score DESC"
        if limit is not None:
            sqlSelect = sqlSelect + " LIMIT " + str(limit)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        record = cur.fetchall()
        return (record)

    def fetch_best_IgorAlignments_By_seq_index(self, strGene, seq_index):
        """
        Fetch IgorAlignments from database by seq_index.
        :param strGene: string to specify the type of gene V, D or J
        :param seq_index: IgorIndexedSequences index
        """
        sqlSelect = "SELECT * FROM Igor" + strGene.upper() + "Alignments WHERE seq_index==" + str(
            seq_index) + " ORDER BY score DESC LIMIT 1"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        record = cur.fetchone()
        return (record)

    def get_best_IgorAlignment_data_By_seq_index(self, strGene, seq_index):
        from .IgorIO import IgorAlignment_data
        best_align_data_record = self.fetch_best_IgorAlignments_By_seq_index(strGene, seq_index)
        # print("best_align_data_record ", best_align_data_record)
        best_align_data = IgorAlignment_data.load_FromSQLRecord(best_align_data_record)
        best_align_data.strGene_class = strGene

        gene_record = self.fetch_IgorGeneTemplate_By_gene_id(best_align_data.strGene_class, best_align_data.gene_id)
        best_align_data.strGene_name = gene_record[1]
        best_align_data.strGene_seq = gene_record[2]
        return best_align_data

    # TODO: Create an IgorAlignment_data instance with a better sql query (join the necessary tables.)
    # TODO:

    def get_IgorAlignment_data_list_query(self, strGene, seq_index, where=None):
        sqlSelect = "SELECT * FROM Igor" + strGene.upper() + "Alignments WHERE seq_index==" + str(seq_index) + ""
        from .IgorIO import IgorAlignment_data
        align_data_records = self.fetch_IgorAlignments_By_seq_index(strGene, seq_index)
        align_data_list = list()
        for align_record in align_data_records:
            align_data = IgorAlignment_data.load_FromSQLRecord(align_record)
            align_data.strGene_class = strGene

            gene_record = self.fetch_IgorGeneTemplate_By_gene_id(align_data.strGene_class, align_data.gene_id)
            align_data.strGene_name = gene_record[1]
            align_data.strGene_seq = gene_record[2]
            align_data_list.append(align_data)
        return align_data_list

    def get_IgorAlignment_data_list_By_seq_index(self, strGene, seq_index, limit=None):
        from .IgorIO import IgorAlignment_data
        align_data_records = self.fetch_IgorAlignments_By_seq_index(strGene, seq_index, limit=limit)
        # best_align_data_record = self.fetch_best_IgorAlignments_By_seq_index(strGene, seq_index)
        # print("best_align_data_record ", align_data_records)
        align_data_list = list()
        for align_record in align_data_records:
            align_data = IgorAlignment_data.load_FromSQLRecord(align_record)
            align_data.strGene_class = strGene

            gene_record = self.fetch_IgorGeneTemplate_By_gene_id(align_data.strGene_class, align_data.gene_id)
            align_data.strGene_name = gene_record[1]
            align_data.strGene_seq = gene_record[2]
            align_data_list.append(align_data)
        return align_data_list

    def get_DataFrame_IgorAlignment_By_seq_index(self, strGene, seq_index, limit=None):
        records = self.fetch_IgorAlignments_By_seq_index(strGene, seq_index, limit=limit)
        pd.DataFrame.from_records(records)

    def appendList_IgorAlignments_data_By_seq_index(self, strGene_class, seq_index, alnDataList=None):
        """
        Append to a list of IgorAlignment_data objects given gene class ("V", "D", "J"), seq_index 
        append a list to append the objects.
        :param strGene_class: string to specify the type of gene V, D or J.
        :param seq_index: IgorIndexedSequences index.
        :param alnDataList: List of IgorAlignment_data objects.
        """
        if alnDataList == None:
            alnDataList = list()
        
        try:
            alignsSQLrecords = self.fetch_IgorAlignments_By_seq_index(strGene_class, seq_index)
            for alignSQLrecord in alignsSQLrecords:
                #print(alignSQLrecord)
                gene_id = alignSQLrecord[1]
                geneTemplateRecord = self.fetch_IgorGeneTemplate_By_gene_id(strGene_class, gene_id)
                strGene_name = geneTemplateRecord[1]
                strGene_seq  = geneTemplateRecord[2]
                aln_data = IgorAlignment_data.IgorAlignment_data.load_FromSQLRecord(alignSQLrecord, strGene_name=strGene_name)
                aln_data.strGene_class = strGene_class
                aln_data.strGene_seq   = strGene_seq
                alnDataList.append(aln_data)
                #print(aln_data.strGene_name, aln_data.score, aln_data.offset, aln_data.insertions)
            return alnDataList
        except Exception as e:
            print(e)
    
    def write_IgorAlignments2Fasta(self, alnDataList):
        print(alnDataList)

    def get_naive_sequence_from_IgorAligment_data(self, seq_index):
        # TODO: use self.dicts and IgorAlignment_data to reconstruct sequence
        # 1. get Indexed Sequence
        indexed_sequence = self.get_IgorIndexedSeq_By_seq_index(seq_index)
        indexed_sequence.offset = 0
        # 2. get Alignments

        # 3. Choose best V and J alignment
        best_v_align_data = self.get_best_IgorAlignment_data_By_seq_index('V', indexed_sequence.seq_index)
        best_j_align_data = self.get_best_IgorAlignment_data_By_seq_index('J', indexed_sequence.seq_index)

        # 4. if D exists then choose the best D gene where offsets are between CDR3 anchors
        d_align_data_list = self.get_IgorAlignment_data_list_By_seq_index('D', indexed_sequence.seq_index)

        # 5. Once D select the highest score.
        # 6. if there is an overlap in V or J segments then, removing that segment

        return ""

    def write_IgorAlignments_to_CSV(self, strGene, flnGeneTemplate, sep=";"):
        print("Exporting alignments to file: ", flnGeneTemplate)

        sqlcmd_select_template = """
            SELECT aln.seq_index, t_gene.gene_name, aln.score, aln.offset, aln.insertions, aln.deletions, 
            aln.mismatches, aln.length, aln.offset_5_p_align, aln.offset_3_p_align
            FROM Igor{upper}Alignments aln 
            INNER JOIN Igor{upper}GeneTemplate t_gene 
            ON aln.{lower}gene_id = t_gene.{lower}gene_id;
            """

        sqlSelect = sqlcmd_select_template.format(upper=strGene.upper(), lower=strGene.lower())
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()

        str_file_header = "seq_index"+sep+"gene_name"+sep+"score"+sep+"offset"+sep+"insertions"+sep+"deletions"+sep+"mismatches"+sep+"length"+sep+"5_p_align_offset"+sep+"3_p_align_offset" + "\n"
        with open(flnGeneTemplate, "w") as ofile:
            ofile.write(str_file_header)
            for record in records:
                csvline = sep.join(map(str, record)).replace("[", "{").replace("]", "}")
                ofile.write(csvline + "\n")

    def delete_IgorAlignments_Tables(self):
        """
        Method to delete IgorAlignments table.
        """
        tables_list = self.get_list_of_tables_with_name('Igor%Alignments')
        for tablename in tables_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)

    # def get_naive_alignment(self, seq_index):
    #     alnDataListV = db.fetch_IgorAlignments_By_seq_index('V', seq_index)
    #     v_align_data = p3.IgorAlignment_data.load_FromSQLRecord(alnDataListV[0])
    #     print(v_align_data.to_dict())
    #     v_gene_seq = db.fetch_IgorGeneTemplate_By_gene_id('V', v_align_data.gene_id)[2]
    #     print(v_gene_seq)

    def load_IgorIndexedCDR3_FromCSV(self, flnIgorIndexedCDR3):
        """
                Insert indexed CDR3 in database from csv igor indexed_seqs file.
                :param conn:
                :param csvline:
                :return:
                """
        self.execute_query(sqlcmd_ct['indexed_CDR3'])

        self.flnIgorIndexedCDR3 = flnIgorIndexedCDR3
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')
            with open(self.flnIgorIndexedCDR3) as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    # print(csvline)
                    self.insert_IgorIndexedCDR3_FromCSVline(cur, csvline)
            cur.execute('COMMIT')
            # self.conn.commit()
        except sqlite3.Error as e:
            print(e)

    def insert_IgorIndexedCDR3_FromCSVline(self, cur, csvline):
        """
        Insert IGoR indexed_sequences on Database flnIgorDB
        :param csvline:
        """
        # seq_index;v_anchor;j_anchor;CDR3nt;CDR3aa
        sql = ''' INSERT INTO IgorIndexedCDR3(seq_index,v_anchor,j_anchor,CDR3,CDR3_aa)
                  VALUES(?,?,?,?,?) '''

        csvline = csvline.replace('\n', '')
        # FIXME: CHANGE THIS ONLY SAVE THE ANCHORS AND NOT THE CDR3 and CDR3_aa
        # data = tuple(csvline.split(";")[0:2]) # to pick just seq_index, v_anchor, j_anchor
        # if len(data) == 2:
        data = tuple(csvline.split(";"))
        # if len(data) == 2:
        try:
            # cur = self.conn.cursor()
            cur.execute(sql, data)
            # self.conn.commit()
        except sqlite3.Error as e:
            print(data)
            print(e)
            pass

    def fetch_IgorIndexedCDR3_By_seq_index(self, seq_index):
        # print(gene_name)
        sqlSelect = "SELECT * FROM IgorIndexedCDR3 WHERE seq_index = " + str(seq_index) + ";"
        # print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        # if strGene == 'D':
        #    print(sqlSelect)
        record = cur.fetchone()
        return (record)

    def write_IgorIndexedCDR3_to_CSV(self, flnIgorIndexedCDR3, sep=';'):
        print("Saving IGoR's indexed CDR3 to file : ", flnIgorIndexedCDR3)
        sqlSelect = "SELECT * FROM IgorIndexedCDR3;"
        # print(sqlSelect)
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        # if strGene == 'D':
        #    print(sqlSelect)
        records = cur.fetchall()

        str_file_header = "seq_index;v_anchor;j_anchor;CDR3;CDR3_aa" + "\n"
        with open(flnIgorIndexedCDR3, "w") as ofile:
            ofile.write(str_file_header)
            for record in records:
                csvline = sep.join(map(str, record)).replace("[", "{").replace("]", "}")
                ofile.write(csvline + "\n")

    def delete_IgorIndexedCDR3_Tables(self):
        """
        Method to delete IgorAlignments table.
        """
        tables_list = self.get_list_of_tables_with_name('IgorIndexedCDR3')
        for tablename in tables_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)

    # def get_IgorIndexedCDR3_By_seq_index(self, seq_index):
    #     record = self.fetch_IgorIndexedCDR3_By_seq_index(seq_index)
    #     return IgorIndexedSequence.load_FromSQLRecord(record)

    ############################
    def load_IgorModel_Parms(self, mdl_parms): #fln_model_parms):
        ################## TODO: Insert Event_list
        self.execute_query(sqlcmd_ct['MP_Event_list'])
        # self.flnIgorModel_Parms = fln_model_parms
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')

            from .IgorIO import IgorModel_Parms
            # mdl_parms = IgorModel_Parms(model_parms_file=fln_model_parms)
            for event in mdl_parms.Event_list:
                event_dict = event.to_dict()
                # print("-"*50, event.nickname)
                event_realizations_list = event_dict.pop('realizations')
                event_dict['realizations_table'] = "IgorER_" + event_dict['nickname']
                self.insert_IgorRec_Event_FromDict(cur, event_dict)
                # Create database with realizations
                script_table = sqlcmd_ct['ER_event_template'].format(event_dict['realizations_table'])
                # insert realizations from realizations object
                try:
                    # Create table of realizations
                    cur.execute(script_table)
                    # Insert realizations
                    for realization in event_realizations_list:
                        # print(realization.to_dict())
                        self.insert_IgorEvent_realization_FromDict(cur, event_dict['realizations_table'], realization.to_dict())
                except sqlite3.Error as e:
                    print("Problem with realization", realization.to_dict())
                    print(e)

            cur.execute('COMMIT')
            # self.conn.commit()
        except sqlite3.Error as e:
            print("Couldn't create IgorModel_Parms table")
            print(e)

        ################## TODO: Insert Edges XXX
        # Create database if not exist.
        self.execute_query(sqlcmd_ct['MP_Edges'])
        cur = self.conn.cursor()
        sql = 'INSERT INTO IgorMP_Edges(parent_event,child_event) VALUES (?,?)'
        cur.execute('BEGIN TRANSACTION')
        for edge in mdl_parms.Edges:
            # Insert in Edges table
            try:
                # cur = self.conn.cursor()
                data = (mdl_parms.dictNameNickname[ edge[0] ], mdl_parms.dictNameNickname[ edge[1] ])
                print(data)
                cur.execute(sql, data)
                # self.conn.commit()
            except sqlite3.Error as e:
                print(edge, data)
                print(e)
                exit()
                # pass
        cur.execute('COMMIT')

        ################## TODO: Insert ErrorRate
        try:
            self.execute_query(sqlcmd_ct['MP_ErrorRate'])
            cur = self.conn.cursor()
            columns = ', '.join(mdl_parms.ErrorRate_dict.keys())
            placeholders = ':' + ', :'.join(mdl_parms.ErrorRate_dict.keys())
            sql = 'INSERT INTO IgorMP_ErrorRate (%s) VALUES (%s)' % (columns, placeholders)
            cur.execute('BEGIN TRANSACTION')
            # print(mdl_parms.ErrorRate_dict)
            cur.execute(sql, mdl_parms.ErrorRate_dict)
            cur.execute('COMMIT')
        except Exception as e:
            print("Couldn't insert ErrorRate in database")
            print(e)

    def insert_IgorRec_Event_FromDict(self, cur, event_dict: dict):
        # print(event_dict.keys())
        columns = ', '.join(event_dict.keys())
        placeholders = ':' + ', :'.join(event_dict.keys())
        sql = 'INSERT INTO IgorMP_Event_list (%s) VALUES (%s)' % (columns, placeholders)
        # print(sql)
        try:
            # cur = self.conn.cursor()
            cur.execute(sql, event_dict)
            # self.conn.commit()
        except sqlite3.Error as e:
            print(event_dict)
            print(e)
            pass

    def insert_IgorEvent_realization_FromDict(self, cur, event_table, event_realization_dict):
        # TODO: IgorEvent_realization
        columns = ', '.join(event_realization_dict.keys())
        placeholders = ':' + ', :'.join(event_realization_dict.keys())
        tmp_sql = 'INSERT INTO '+event_table+' (%s) VALUES (%s)'
        sql = tmp_sql % (columns, placeholders)
        # print(sql)
        try:
            # cur = self.conn.cursor()
            cur.execute(sql, event_realization_dict)
            # self.conn.commit()
        except sqlite3.Error as e:
            print(event_realization_dict)
            print(e)
            pass

    def load_IgorModel_Marginals(self, mdl_xdata:dict):
        for event_nickname in mdl_xdata.keys():
            da = mdl_xdata[event_nickname]
            lista = list(dict(da.coords).keys())
            # print(lista)
            for dim in da.dims:
                lista.remove(dim)
            da = da.drop(lista)
            # Create table of not real marginals
            if da.attrs['event_type'] == 'DinucMarkov':
                self.execute_query( sqlcmd_ct_Model_Marginals_DinucMarkov(event_nickname, ['x', 'y']) )
                cur = self.conn.cursor()
                try:
                    cur.execute('BEGIN TRANSACTION')
                    df = da.to_dataframe(name='P')
                    df.reset_index(inplace=True)
                    # Rename names of dataframe
                    dicto2rename = dict()
                    for col in df.columns:
                        if not col == 'P':
                            dicto2rename[col] = "id_" + col
                    df.rename(dicto2rename, axis='columns', inplace=True)

                    mm_table = "IgorMM_" + event_nickname
                    columns_list = list(df.columns)
                    str_columns = ",".join(columns_list)
                    placeholders = ':' + ', :'.join(columns_list)
                    sql = 'INSERT INTO ' + mm_table + ' (%s) VALUES (%s);' % (str_columns, placeholders)
                    #print("sql : ", sql)
                    for index, row in df.iterrows():
                        cur.execute(sql, row.to_dict())

                    cur.execute('COMMIT')
                except sqlite3.Error as e:
                    print("-" * 40)
                    print("ERROR: Can't insert in " + event_nickname + " table")
                    print(e)
                    print("-"*40)

            else:
                tmp_list = [event_nickname] + da.attrs['parents']
                da = da.transpose(*tmp_list)
                # Create table for model marginals
                self.execute_query( sqlcmd_ct_Model_Marginals(event_nickname, da.attrs['parents']) )

                cur = self.conn.cursor()
                try:
                    cur.execute('BEGIN TRANSACTION')
                    df = da.to_dataframe(name='P')
                    df.reset_index(inplace=True)

                    # Rename names of dataframe
                    dicto2rename = dict()
                    for col in df.columns:
                        if not col == 'P':
                            dicto2rename[col] = "id_"+col
                    df.rename(dicto2rename, axis='columns', inplace=True)

                    mm_table = "IgorMM_"+event_nickname
                    columns_list = list(df.columns)
                    str_columns = ",".join(columns_list )
                    placeholders = ':' + ', :'.join(columns_list)
                    sql = 'INSERT INTO ' + mm_table + ' (%s) VALUES (%s)' % (str_columns, placeholders)

                    for index, row in df.iterrows():
                        cur.execute(sql, row.to_dict())
                    cur.execute('COMMIT')
                    # self.conn.commit()
                except sqlite3.Error as e:
                    print("-" * 40)
                    print("ERROR: Can't insert in "+event_nickname+" table")
                    print(e)
                    print("-" * 40)

    def insert_IgorModel_Marginals_FromDict(self, cur, event_table, event_realization_dict):
        # TODO: IgorEvent_realization
        columns = ', '.join(event_realization_dict.keys())
        placeholders = ':' + ', :'.join(event_realization_dict.keys())
        tmp_sql = 'INSERT INTO '+event_table+' (%s) VALUES (%s)'
        sql = tmp_sql % (columns, placeholders)
        # print(sql)
        try:
            # cur = self.conn.cursor()
            cur.execute(sql, event_realization_dict)
            # self.conn.commit()
        except sqlite3.Error as e:
            print(event_realization_dict)
            print(e)
            pass

    def load_IgorModel_FromTXT(self, flnIgorModel_Parms, flnIgorModel_Marginals):
        self.flnIgorModel_Parms = flnIgorModel_Parms
        self.flnIgorModel_Marginals = flnIgorModel_Marginals
        from .IgorIO import IgorModel
        mdl = IgorModel(model_parms_file=flnIgorModel_Parms, model_marginals_file=flnIgorModel_Marginals)
        self.load_IgorModel(self, mdl)

    def load_IgorModel(self, mdl):
        print("Loading parms to database")
        self.load_IgorModel_Parms(mdl.parms)
        print("Loading marginals to database")
        self.load_IgorModel_Marginals(mdl.xdata)

    def write_IgorModel_to_TXT(self, flnIgorModel_Parms, flnIgorModel_Marginals):
        print("Saving IGoR's model parms to file: ", flnIgorModel_Parms, " and model marginals to file: ", flnIgorModel_Marginals)
        mdl_parms = self.get_IgorModel_Parms()
        mdl_parms.write_model_parms(flnIgorModel_Parms)
        mdl_marginals = self.get_IgorModel_Marginals()
        mdl_marginals.write_model_marginals(flnIgorModel_Marginals, mdl_parms)

    def write_IgorModel_Parms_to_TXT(self, flnIgorModel_Parms):
        # from .IgorIO import IgorModel_Parms
        print("Saving IGoR's model parms to file: ", flnIgorModel_Parms) #
        mdl_parms = self.get_IgorModel_Parms()
        mdl_parms.write_model_parms(flnIgorModel_Parms)

    def write_IgorModel_Marginals_to_TXT(self, flnIgorModel_Marginals):
        from .IgorIO import IgorModel_Parms, IgorModel_Marginals
        print("Saving IGoR's and model marginals to file: ", flnIgorModel_Marginals)
        mdl_parms = self.get_IgorModel_Parms()
        mdl_marginals = self.get_IgorModel_Marginals()
        mdl_marginals.write_model_marginals(flnIgorModel_Marginals, mdl_parms)


    def delete_IgorModel_Tables(self):
        for tabla in sql_tablename_patterns_dict['model']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
                counts = self.execute_query(sql_cmd_drop)

    def delete_IgorModel_Parms_Tables(self):
        for tabla in ['IgorMP_%', 'IgorER_%']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
                counts = self.execute_query(sql_cmd_drop)

    def delete_IgorModel_Marginals_Tables(self):
        for tabla in ['IgorMM_%']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
                counts = self.execute_query(sql_cmd_drop)


    def load_IgorBestScenarios_FromCSV(self, flnIgorBestScenarios, mdl=None):
        if mdl is None:
            mdl = self.get_IgorModel()

        self.flnIgorBestScenarios = flnIgorBestScenarios
        with open(self.flnIgorBestScenarios, "r") as fp:
            #### CREATE DATABASE
            ## Get best scenarios header file
            igor_bsfileHeader = fp.readline().replace('\n', '').replace('Mismatches', 'mismatches')
            file_header_list = igor_bsfileHeader.split(';')
            bs_events_name_list = file_header_list[3:-1]
            ## Change event_name to event_nickname using dictionary.
            # Get dictionary to change names to nicknames
            events_name_nickname_dict = mdl.parms.get_event_dict('name', 'nickname')
            print("events_name_nickname_dict", events_name_nickname_dict)
            bs_events_nickname_list = [ events_name_nickname_dict[event_name] for event_name in bs_events_name_list]

            # Now generate sql_id, sql_type and foreign_key in the file order
            events_dict_list = list()
            sql_list_ct_table_cols = list()
            sql_list_ct_table_foreign_keys = list()
            sql_list_ins_table = list()
            self.sql_list_of_listtypes = list()
            for ii, event_nickname in enumerate(bs_events_nickname_list) :
                event = mdl.parms.get_Event(event_nickname)
                event_dict = event.to_dict()
                del event_dict['realizations']
                if event.event_type == 'DinucMarkov':
                    event_dict['sql_id'] = event.nickname
                    event_dict['sql_type'] = "text"
                    sql_list_ct_table_cols.append(event_dict['sql_id']+" "+event_dict['sql_type'])
                    sql_list_ins_table.append(event_dict['sql_id'])
                    self.sql_list_of_listtypes.append(3+ii)
                else:
                    event_dict['sql_id'] = 'id_'+event.nickname
                    event_dict['sql_type'] = "integer"
                    event_dict['sql_foreign_key'] = "FOREIGN KEY (id_{}) REFERENCES IgorER_{} (id)".format(event_nickname, event_nickname)
                    sql_list_ct_table_cols.append(event_dict['sql_id'] + " " + event_dict['sql_type'])
                    sql_list_ins_table.append(event_dict['sql_id'])
                    sql_list_ct_table_foreign_keys.append(event_dict['sql_foreign_key'])
                events_dict_list.append(event_dict)

            self.sql_list_of_listtypes.append(3+len(bs_events_nickname_list)) # mismatches position

            # Now make script to create table in database and create table
            self.execute_query(
                sqlcmd_ct['BestScenario_template'].format(','.join(sql_list_ct_table_cols),
                                                          ','.join(sql_list_ct_table_foreign_keys))
            )

            ### INSERT VALUES
            sql_cols_list =['seq_index', 'scenario_rank', 'scenario_proba_cond_seq'] + sql_list_ins_table + ['mismatches', 'mismatcheslen']
            sql_str_cols = ','.join(sql_cols_list)
            sql_str_aux = ','.join(['?' for col in sql_cols_list])

            self.sql_BestScenarios_cols_list = sql_cols_list
            self.sqlcmd_ins_bs = ''' INSERT INTO IgorBestScenarios({}) VALUES({}) '''.format(sql_str_cols, sql_str_aux)

            #print(self.sqlcmd_ins_bs)

            # INSERT VALUES IN DATABASE
            cur = self.conn.cursor()
            try:
                cur.execute('BEGIN TRANSACTION')
                with open(self.flnIgorBestScenarios) as fp:
                    csvline = fp.readline()
                    while csvline:
                        csvline = fp.readline()
                        self.insert_IgorBestScenarios_FromCSVline(cur, csvline)
                cur.execute('COMMIT')
                # self.conn.commit()
            except sqlite3.Error as e:
                print(e)

    def insert_IgorBestScenarios_FromCSVline(self, cur, csvline):
        # sql = ''' INSERT INTO IgorBestScenarios({}) VALUES({}) '''
        csvline = csvline.replace('\n', '')
        csvlist = csvline.split(";")
        # FIXME: FIND A BETTER WAY TO REPLACE THE TYPE FOR INSERTION
        for ii in range(len(csvlist)):
            if ii in self.sql_list_of_listtypes:
                csvlist[ii] = csvlist[ii].replace("(", "[").replace(")", "]")  # lists vd_dinucl, dj_dinucl, mismatches
            else:
                csvlist[ii] = csvlist[ii].replace("(", "").replace(")", "")

        # print(csvlist)

        # data[-1] convert it to list and
        try:
            data = csvlist + [len(eval(csvlist[-1]))]
            try:
                # cur = self.conn.cursor()
                cur.execute(self.sqlcmd_ins_bs, data)
                # self.conn.commit()
            except sqlite3.Error as e:
                print(data)
                print(e)
                pass
        except Exception as e:
            print("ERROR: insert_IgorBestScenarios_FromCSVline")
            print(csvlist)
            print(e)
            pass

    def write_IgorBestScenarios_to_CSV(self, flnIgorBestScenarios, mdl=None, sep=';'):
        """
        seq_index;scenario_rank;scenario_proba_cond_seq;GeneChoice_V_gene_Undefined_side_prio7_size168;GeneChoice_J_gene_Undefined_side_prio7_size16;GeneChoice_D_gene_Undefined_side_prio6_size3;Deletion_V_gene_Three_prime_prio5_size21;Deletion_D_gene_Five_prime_prio5_size21;Deletion_D_gene_Three_prime_prio5_size21;Deletion_J_gene_Five_prime_prio5_size21;Insertion_VD_genes_Undefined_side_prio4_size41;DinucMarkov_VD_genes_Undefined_side_prio3_size16;Insertion_DJ_gene_Undefined_side_prio2_size41;DinucMarkov_DJ_gene_Undefined_side_prio1_size16;Mismatches
        99;1;0.0401246;(109);(12);(2);(9);(16);(4);(7);(5);(2,2,0,3,0);(5);(0,2,0,3,3);()
        99;2;0.0401246;(109);(12);(0);(9);(9);(7);(7);(5);(2,2,0,3,0);(5);(0,2,0,3,3);()
        99;3;0.0100351;(109);(12);(2);(10);(13);(7);(7);(0);();(11);(0,2,0,3,3,2,2,2,0,0,3);()
        99;4;0.0100351;(109);(12);(2);(9);(14);(7);(7);(0);();(11);(0,2,0,3,3,2,2,2,0,0,3);()
        """
        print("Saving IGoR's best scenarios to file: ", flnIgorBestScenarios)
        if mdl is None:
            mdl = self.get_IgorModel()

        ### MAKE THE HEADER with name using nicknames
        # Transform the nicknames with id
        db_events_nickname_name_dict = dict()
        for event in mdl.parms.Event_list:
            if event.event_type == 'DinucMarkov':
                db_events_nickname_name_dict[event.nickname] = event.name
            else:
                db_events_nickname_name_dict["id_"+event.nickname] = event.name
        db_events_nickname_name_dict['mismatches'] = 'Mismatches'
        print(db_events_nickname_name_dict)

        list_bs_columns_db = [el[0] for el in self.get_columns_type_of_tables('IgorBestScenarios')[:-1]]
        tmp_list = list()
        for aaa in list_bs_columns_db:
            if aaa in db_events_nickname_name_dict.keys():
                tmp_list.append(db_events_nickname_name_dict[aaa])
            else:
                tmp_list.append(aaa)
        # str_file_header
        # str_file_header = sep.join(list_bs_columns_db)
        str_file_header = sep.join(tmp_list)+"\n"

        str_sql_columns_to_query = ",".join(list_bs_columns_db)
        sqlSelect = "SELECT " + str_sql_columns_to_query + " FROM IgorBestScenarios;"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()

        # str_file_header = "seq_index;gene_name;score;offset;insertions;deletions;mismatches;length;5_p_align_offset;3_p_align_offset" + "\n"
        with open(flnIgorBestScenarios, "w") as ofile:
            ofile.write(str_file_header)
            for record in records:
                csvline = sep.join(map(str, record)).replace("[", "{").replace("]", "}")
                ofile.write(csvline + "\n")

    def delete_IgorBestScenarios_Tables(self):
        tablenames_list = self.get_list_of_tables_with_name('IgorBestScenarios')
        for tablename in tablenames_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)

    def export_IgorBestScenarios_to_AIRR(self, flnAIRR_arrangement, mdl=None, sep='\t'):
        """Export Igor best scenarios to airr rearrangement format"""
        if mdl is None:
            mdl = self.get_IgorModel()

        from .AIRR import AIRR_VDJ_rearrangement
        import airr

        # FIXME: IF VDJ THEN:
        airr_fields = AIRR_VDJ_rearrangement.list_of_fields()
        airr_rearrangement_writer = airr.create_rearrangement(flnAIRR_arrangement, fields=airr_fields)
        for seq_index, sequence in self.fetch_IgorIndexedSeq_records():
            scenarios_list = self.get_IgorBestScenarios_By_seq_index_IgorModel(seq_index, mdl)
            for scenario in scenarios_list:
                v_best_aln = self.get_best_IgorAlignment_data_By_seq_index("V", seq_index)

                airr_rearrangement_dict = mdl.get_AIRR_VDJ_rearragement_dict_from_scenario(scenario, sequence, v_offset=v_best_aln.offset)
                airr_rearrangement_dict['scenario_rank'] = scenario.scenario_rank
                airr_rearrangement_dict['scenario_proba_cond_seq'] = scenario.scenario_proba_cond_seq
                airr_rearrangement_dict['pgen'] = self.fetch_IgorPgen_By_seq_index(seq_index)[1]
                cdr3_record = self.fetch_IgorIndexedCDR3_By_seq_index(seq_index)
                airr_rearrangement_dict['junction'] = cdr3_record[3]
                airr_rearrangement_dict['junction_aa'] = cdr3_record[4]
                airr_rearrangement_writer.write(airr_rearrangement_dict)

        airr_rearrangement_writer.close()




        # n_d_5_del = self.mdlParms.Event_dict[strEv].loc[self.id_d_5_del]['value']
        # name_D = self.mdlParms.Event_dict[strEv].loc[self.id_d_gene]['name']
        # v_call d_call c_call v_score d_score j_score v_cigar j_evalue v_identity v_start v_end v_germ_start v_germ_end np1_seq np1_length np2_seq np2_length

        # from pygor3.IgorIO import IgorScenario
        # # OPEN FILE
        # for seq_index in self.fetch_IgorIndexedSeq_indexes():
        #     scenarios = self.get_IgorBestScenarios_By_seq_index(seq_index)
        #     sequence = self.fetch_IgorIndexedSeq_By_seq_index(seq_index)
        #     for scenario in scenarios:
        #         scenario.export_to_AIRR_line()
        #
        #
        # ### MAKE THE HEADER with name using nicknames
        # # Transform the nicknames with id
        # db_events_nickname_name_dict = dict()
        # for event in mdl.parms.Event_list:
        #     if event.event_type == 'DinucMarkov':
        #         db_events_nickname_name_dict[event.nickname] = event.name
        #     else:
        #         db_events_nickname_name_dict["id_" + event.nickname] = event.name
        # db_events_nickname_name_dict['mismatches'] = 'Mismatches'
        # print(db_events_nickname_name_dict)
        #
        # list_bs_columns_db = [el[0] for el in self.get_columns_type_of_tables('IgorBestScenarios')[:-1]]
        # tmp_list = list()
        # for aaa in list_bs_columns_db:
        #     if aaa in db_events_nickname_name_dict.keys():
        #         tmp_list.append(db_events_nickname_name_dict[aaa])
        #     else:
        #         tmp_list.append(aaa)
        # # str_file_header
        # # str_file_header = sep.join(list_bs_columns_db)
        # str_file_header = sep.join(tmp_list) + "\n"
        #
        # str_sql_columns_to_query = ",".join(list_bs_columns_db)
        # sqlSelect = "SELECT " + str_sql_columns_to_query + " FROM IgorBestScenarios;"
        # cur = self.conn.cursor()
        # cur.execute(sqlSelect)
        # records = cur.fetchall()
        #
        # # str_file_header = "seq_index;gene_name;score;offset;insertions;deletions;mismatches;length;5_p_align_offset;3_p_align_offset" + "\n"
        # with open(flnIgorBestScenarios, "w") as ofile:
        #     ofile.write(str_file_header)
        #     for record in records:
        #         csvline = sep.join(map(str, record)).replace("[", "{").replace("]", "}")
        #         ofile.write(csvline + "\n")



    ###### PGEN
    def load_IgorPgen_FromCSV(self, flnIgorPgen):
        print(flnIgorPgen)
        self.execute_query(sqlcmd_ct['Pgen'])
        self.flnIgorPgen = flnIgorPgen
        cur = self.conn.cursor()
        try:
            cur.execute('BEGIN TRANSACTION')
            with open(self.flnIgorPgen) as fp:
                csvline = fp.readline()
                while csvline:
                    csvline = fp.readline()
                    # print(csvline)
                    self.insert_load_IgorPgen_FromCSVline(cur, csvline)
            cur.execute('COMMIT')
            # self.conn.commit()
        except sqlite3.Error as e:
            print(e)

    def insert_load_IgorPgen_FromCSVline(self, cur, csvline):
        sql = ''' INSERT INTO IgorPgen(seq_index,Pgen_estimate)
                          VALUES(?,?) '''
        csvline = csvline.replace('\n', '')
        data = tuple(csvline.split(";"))
        try:
            cur.execute(sql, data)
        except sqlite3.Error as e:
            print(data)
            print(e)
            pass

    def fetch_IgorPgen(self):
        sqlSelect = "SELECT * FROM IgorPgen;"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        return records

    def fetch_IgorPgen_By_seq_index(self, seq_index):
        sqlSelect = "SELECT * FROM IgorPgen WHERE seq_index = " + str(seq_index) + ";"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        record = cur.fetchone()
        return record

    def write_IgorPgen_to_CSV(self, flnIgorPgen, sep=';'):
        print("Writing Pgen to file: ", flnIgorPgen)
        str_header = "seq_index"+sep+"Pgen_estimate\n"
        with open(flnIgorPgen, "w") as ofile:
            # seq_index,Pgen_estimate
            ofile.write(str_header)
            for record in self.fetch_IgorPgen():
                try:
                    ofile.write(str(record[0])+sep+str(record[1])+"\n" )
                except Exception as e:
                    print("ERROR: write_IgorPgen_to_CSV ", e)

    def delete_IgorPgen_Tables(self):
        tablenames_list = self.get_list_of_tables_with_name('IgorPgen')
        for tablename in tablenames_list:
            sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
            self.execute_query(sql_cmd_drop)


    ###### return IGoR Model
    def get_IgorModel(self):
        from .IgorIO import IgorModel
        mdl = IgorModel.load_from_parms_marginals_object( self.get_IgorModel_Parms(), self.get_IgorModel_Marginals() )
        return mdl

    def get_IgorModel_Marginals(self):
        print("-"*5, "Marginals", "-"*5)
        from .IgorIO import IgorModel_Marginals
        marginals = IgorModel_Marginals()

        # So there are two different ways to implement this.
        # 1. use a model parms to get all the nicknames and explore the name tables
        # 2. Check query in db the nicknames
        # 3. How to get all the tables with the prefix "IgorMM_"
        # And the best way to do this is to ask the database (2)
        # Get all the list of nicknames:
        nickname_list = self.execute_select_query("SELECT nickname FROM IgorMP_Event_list;")
        nickname_list = [nickname[0] for nickname in nickname_list]

        marginals.network_dict = dict()
        marginals.marginals_dict = dict()

        for event_nickname in nickname_list:
            col_names = self.get_table_colsname_list("IgorMM_"+event_nickname)
            records = self.execute_select_query("SELECT * FROM IgorMM_"+event_nickname+";")

            print(event_nickname) # , records)
            events_tmp = [x[3:] for x in col_names if x != 'P']
            marginals.network_dict[event_nickname] = events_tmp
            df = pd.DataFrame(np.array(records), columns=events_tmp+['P'])
            dimensions = list()
            for event_name in events_tmp:
                dim_ev = len(df[event_name].unique())
                dimensions.append(dim_ev)

            marginals.marginals_dict[event_nickname] = df['P'].to_numpy().reshape(dimensions)

        return marginals

    def get_table_colsname_list(self, tablename):
        sqlcmd_qry = sqlcmd_template_qry_cols.format(tablename=tablename)
        col_names = self.execute_select_query(sqlcmd_qry)
        return [colname[0] for colname in col_names]

    def get_IgorModel_Parms(self):
        from .IgorIO import IgorModel_Parms
        mdl_parms = IgorModel_Parms()
        # tb_IgorMP_Event_list_columns = self.execute_select_query(
        #     "SELECT name FROM pragma_table_info('IgorMP_Event_list');")
        # tb_IgorMP_Event_list_columns = [aa[0] for aa in tb_IgorMP_Event_list_columns]
        # event_type, seq_type, seq_side, priority, nickname

        mdl_parms.Event_list = self.get_Event_list()
        dict_nickname_name = mdl_parms.get_event_dict('nickname', 'name')
        mdl_parms.Edges = list()
        for edge in self.get_Edges():
            mdl_parms.Edges.append( [dict_nickname_name[edge[0]], dict_nickname_name[edge[1]] ])

        mdl_parms.ErrorRate_dict = self.get_ErrorRate_dict()

        mdl_parms.gen_EventDict_DataFrame()
        return mdl_parms

    def get_Event_list(self):
        from .IgorIO import IgorRec_Event
        from .IgorIO import IgorEvent_realization
        sql_cmd = "SELECT event_type, seq_type, seq_side, priority, nickname, realizations_table FROM IgorMP_Event_list;"
        Event_list = list()
        for rec in self.execute_select_query(sql_cmd):
            event = IgorRec_Event(*rec[:-1])
            str_realization_table = rec[-1]
            realization_dbrecords = self.execute_select_query("SELECT * FROM " + str_realization_table + ";")
            for realization_dbrec in realization_dbrecords:
                realization = IgorEvent_realization()
                if event.event_type == "GeneChoice":
                    realization.id = int(realization_dbrec[0])
                    realization.value = realization_dbrec[1]
                    realization.name = realization_dbrec[2]
                elif event.event_type == "DinucMarkov":
                    realization.id = int(realization_dbrec[0])
                    realization.value = realization_dbrec[1]
                else:
                    realization.id = int(realization_dbrec[0])
                    realization.value = int(realization_dbrec[1])
                event.add_realization(realization)
            #print(event)
            Event_list.append(event)

        return Event_list

    def get_Edges(self):
        sql_cmd = "SELECT parent_event, child_event FROM IgorMP_Edges;"
        Edges = self.execute_select_query(sql_cmd)
        return Edges

    def get_ErrorRate_dict(self):
        ErrorRate_dict = dict()
        sql_cmd = "SELECT error_type, error_values FROM IgorMP_ErrorRate;"
        ErrorRate_record = self.execute_select_query(sql_cmd)[0]
        print("ErrorRate_record : ", ErrorRate_record)
        ErrorRate_dict['error_type'] = ErrorRate_record[0]
        ErrorRate_dict['error_values'] = ErrorRate_record[1]
        return ErrorRate_dict

    def get_list_of_tables_with_name(self, table_name_pattern):
        cmd_qry = """
        SELECT name FROM sqlite_master WHERE type ='table' AND name NOT LIKE 'sqlite_%'
	    AND name LIKE '{tablename}';
        """
        sql_cmd_qry = cmd_qry.format(tablename=table_name_pattern)
        records = self.execute_select_query(sql_cmd_qry)
        tables_list = [record[0] for record in records]
        return tables_list

    def get_dict_of_Igortablename_sql(self):
        cmd_qry = """
        SELECT name, sql FROM sqlite_master WHERE type ='table' AND name NOT LIKE 'sqlite_%'
        AND name LIKE 'Igor%';
        """
        records = self.execute_select_query(cmd_qry)
        tablename_sql_dict = dict()
        for record in records:
            tablename_sql_dict[record[0]] = record[1]

        return tablename_sql_dict

    def delete_IgorModel_Tables(self):
        tablepatterns_list = ["IgorER_%", "IgorMM_%", "IgorMP_%"]
        for tablepattern in tablepatterns_list:
            tablenames_list = self.get_list_of_tables_with_name(tablepattern)
            for tablename in tablenames_list:
                sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
                self.execute_query(sql_cmd_drop)

    #########
    def get_columns_type_of_tables(self, tablename):
        select_qry = "SELECT name, type FROM pragma_table_info('{tablename}');".format(tablename=tablename)
        return self.execute_select_query(select_qry)

    def gen_IgorBestScenarios_cols_list(self):
        self.sql_IgorBestScenarios_cols_name_type_list = self.execute_select_query("SELECT name, type FROM pragma_table_info('IgorBestScenarios');")

    def fetch_IgorBestScenarios_By_events_dict(self, events_id_list_dict):
        print(events_id_list_dict)
        sqlselect = "SELECT * FROM IgorBestScenarios"
        sql_tmp_ANDs_list = list()
        print(sqlselect)
        for key, values in events_id_list_dict.items():
            print(key, values)
            sql_tmp_ANDs_list.append( key+ " IN "+ "(" + ",".join(map(str,values))+ ")" )

        if len(sql_tmp_ANDs_list) == 0:
            sql_where = ";"
        else:
            sql_where = " WHERE " +  " AND ".join(sql_tmp_ANDs_list) + ";"

        return self.execute_select_query(sqlselect + sql_where)

    def fetch_IgorBestScenarios_By_seq_index(self, seq_index):
        sqlSelect = "SELECT * FROM IgorBestScenarios WHERE seq_index==" + str(
            seq_index) #+ " ORDER BY score DESC LIMIT 1"
        cur = self.conn.cursor()
        cur.execute(sqlSelect)
        records = cur.fetchall()
        return (records)

    def get_IgorBestScenariosDataframe_By_seq_index(self, seq_index):
        bs_records = self.fetch_IgorBestScenarios_By_seq_index(seq_index)
        import pandas as pd
        self.gen_IgorBestScenarios_cols_list()
        columnas = [ item[0] for item in self.sql_IgorBestScenarios_cols_name_type_list]
        return pd.DataFrame.from_records(bs_records, columns=columnas)


    def get_IgorBestScenarios_By_seq_index(self, seq_index):
        from .IgorIO import IgorScenario
        # FIXME: THIS METHOD SHOULD BE EVALUATED ONCE AFTER DATA IS LOADED WITH A TRY IF TABLE EXIST
        #  TO AVOID THE UNNECESSARY RERUNNING.
        self.gen_IgorBestScenarios_cols_list()
        scenarios_list = list()
        for record in self.fetch_IgorBestScenarios_By_seq_index(seq_index):
            scenario = IgorScenario.load_FromSQLRecord(record, self.sql_IgorBestScenarios_cols_name_type_list)
            scenarios_list.append(scenario)
        return scenarios_list

    # FIXME : FIND A BETTER DESING FOR THIS FUNCTION
    def get_IgorBestScenarios_By_seq_index_IgorModel(self, seq_index, mdl):
        scenarios_mdl_list = list()
        scenarios_list = self.get_IgorBestScenarios_By_seq_index(seq_index)
        for scen in scenarios_list:
            scen_dict = scen.realizations_ids_dict
            for event in mdl.parms.Event_list:
                if not (event.event_type == 'DinucMarkov'):
                    scen_dict[event.nickname] = scen_dict.pop('id_' + event.nickname)
            scen.realizations_ids_dict = scen_dict
            scenarios_mdl_list.append(scen)

        return scenarios_mdl_list


    def fetch_AIRR_arrangement_By_seq_index(self, seq_index, mdl=None):
        sqlSelect = "SELECT * FROM IgorBestScenarios WHERE seq_index==" + str(
            seq_index) #+ " ORDER BY score DESC LIMIT 1"

        from .IgorIO import IgorScenario

        # TODO: NOW AIRR SCENARIOS
        # 1. Load model used for scenarios.
        if mdl is None:
            mdl = self.get_IgorModel()

        change_dicto = dict()
        for event in mdl.parms.Event_list:
            if event.event_type == 'DinucMarkov':
                change_dicto[event.nickname] = event.nickname
            else:
                change_dicto['id_'+event.nickname] = event.nickname

        # 2. Get the list of columns
        scenarios_col_list = self.get_columns_type_of_tables('IgorBestScenarios')
        scenarios_db_list = list()
        # So scenarios in database have ids, so for use the model I just want the realizations dont care about the rest
        evento = mdl.parms.get_Event(event.nickname)
        from .IgorIO import IgorRec_Event
        evento = IgorRec_Event()
        evento.realizations[realization_id]
        # for the scenarios I only have
        from collections import namedtuple



        for col, tipo in scenarios_col_list:
            print(col, tipo)
            #check in colname in mdl
            if col in change_dicto:
                change_dicto[col]

        scenarios_list = self.get_IgorBestScenarios_By_seq_index_IgorModel()
        # For each scenario in list pass it to airr
        # genomics_dict = self.get_IgorGenomicDataFrame_dict()
        # # 3. Make a map using the column list with model
        #
        # sqlSelect_VDJ= """
        #     SELECT Tscenarios.seq_index as sequence_id,
        #         Tscenarios.scenario_rank,
        #         Tscenarios.scenario_proba_cond_seq,
        #         Tindexed.sequence,
        #         T_v_choice.name as v_call,
        #         T_d_gene.name as d_call,
        #         T_j_choice.name as j_call
        #     FROM IgorBestScenarios Tscenarios
        #         JOIN IgorIndexedSeq Tindexed,
        #             IgorER_v_choice T_v_choice,
        #             IgorER_d_gene T_d_gene,
        #             IgorER_j_choice T_j_choice
        #     WHERE Tindexed.seq_index==Tscenarios.seq_index
        #         AND Tindexed.seq_index == 0
        #         AND Tscenarios.id_v_choice == T_v_choice.id
        #         AND Tscenarios.id_d_gene == T_d_gene.id
        #         AND Tscenarios.id_j_choice == T_j_choice.id;
        #     """
        #
        #
        # cur = self.conn.cursor()
        # cur.execute(sqlSelect)
        # records = cur.fetchall()
        # return (records)

    def calc_IgorBestScenarios_average_of(self, scenario_function, indices_list=None):
        from .IgorIO import IgorScenario
        self.gen_IgorBestScenarios_cols_list()
        if indices_list is None:
            indices_list = self.fetch_IgorIndexedSeq_indexes()
        print("len: ", len(indices_list))
        function_average = 0
        for indx in indices_list:
            # for indx in indexes_list:
            # aln_data = db.get_IgorAlignment_data_list_By_seq_index('V', indx)
            bs_data_list = self.fetch_IgorBestScenarios_By_seq_index(indx)
            tmp_average = 0
            scenario_norm_factor = 0
            for bs_data in bs_data_list:
                bs = IgorScenario.load_FromSQLRecord(bs_data, self.sql_IgorBestScenarios_cols_name_type_list)
                tmp_average = tmp_average + bs.scenario_proba_cond_seq * scenario_function(bs)
                scenario_norm_factor = scenario_norm_factor + bs.scenario_proba_cond_seq
            # FIXME: PERFORMANCE WILL INCREASE IF WE MAKE THE NORMALIZATION BEFORE?
            tmp_average = tmp_average / scenario_norm_factor
            # print("scenario_norm_factor: ", scenario_norm_factor)
            function_average = function_average + tmp_average

        return (function_average / (len(indices_list)))

    # FIXME: What should be the correct method to calculate the average?
    def FIXME_calc_IgorBestScenarios_average_of(self, scenario_function):
        from .IgorIO import IgorScenario
        self.gen_IgorBestScenarios_cols_list()
        indexes_list = self.fetch_IgorIndexedSeq_indexes()
        print("len: ", len(indexes_list))
        function_average = 0
        seq_pgen_tuple_list = self.fetch_IgorPgen()

        pgen_normalization = 0
        for indx, pgen in seq_pgen_tuple_list:
            # for indx in indexes_list:
            # aln_data = db.get_IgorAlignment_data_list_By_seq_index('V', indx)
            bs_data_list = self.fetch_IgorBestScenarios_By_seq_index(indx)
            tmp_average = 0
            scenario_norm_factor = 0
            pgen_normalization = pgen_normalization + pgen
            for bs_data in bs_data_list:
                bs = IgorScenario.load_FromSQLRecord(bs_data, self.sql_IgorBestScenarios_cols_name_type_list)
                tmp_average = tmp_average + bs.scenario_proba_cond_seq * scenario_function(bs)
                scenario_norm_factor = scenario_norm_factor + bs.scenario_proba_cond_seq
            tmp_average = tmp_average / scenario_norm_factor
            # function_average = function_average + pgen * tmp_average
            function_average = function_average + tmp_average

        return (function_average / (pgen_normalization * len(indexes_list)))

    def calc_IgorBestScenarios_sum_of(self, scenario_function):
        from .IgorIO import IgorScenario
        self.gen_IgorBestScenarios_cols_list()
        indexes_list = self.fetch_IgorIndexedSeq_indexes()
        print("len: ", len(indexes_list))
        function_sum = 0
        for indx in indexes_list:
            bs_data_list = self.fetch_IgorBestScenarios_By_seq_index(indx)
            tmp_average = 0
            for bs_data in bs_data_list:
                bs = IgorScenario.load_FromSQLRecord(bs_data, self.sql_IgorBestScenarios_cols_name_type_list)
                tmp_average = tmp_average + scenario_function(bs)
            function_sum = function_sum + tmp_average
            # function_average = function_average + tmp_average

        return (function_sum )


    ################## DATABASE UTILITIES ##################
    def Q_sequences_in_db(self):
        flag_db = False
        for tabla in sql_tablename_patterns_dict['read_seqs']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True

        return flag_db

    def Q_ref_genome_in_db(self):
        flag_db = False
        for tabla in sql_tablename_patterns_dict['ref_genome']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True

        return flag_db

    def Q_ref_genome_in_db_by_gene(self, strGene):
        flag_table = False
        tables_list = self.get_list_of_tables_with_name('Igor'+strGene+'GeneTemplate')
        if len(tables_list) == 1 : flag_table = True
        return flag_table

    def Q_align_in_db(self):
        flag_db = False
        for tabla in sql_tablename_patterns_dict['align']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True

        return flag_db

    def Q_align_in_db_by_gene(self, strGene):
        flag_table = False
        tables_list = self.get_list_of_tables_with_name('Igor' + strGene + 'Alignments')
        if len(tables_list) == 1: flag_table = True
        return flag_table


    def Q_model_in_db(self):
        flag_db = False
        for tabla in sql_tablename_patterns_dict['model']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True

        return flag_db

    def Q_output_in_db(self):
        flag_db = False
        for tabla in sql_tablename_patterns_dict['output']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True

        return flag_db

    def Q_IgorPgen_in_db(self):
        flag_db = False
        for tabla in ['IgorPgen']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True
        return flag_db

    def Q_IgorBestScenarios_in_db(self):
        flag_db = False
        for tabla in ['IgorBestScenarios']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            # print(tables_list)
            if len(tables_list) > 0: flag_db = True
        return flag_db

    def list_from_db(self):
        print("=== Sequences tables igor-reads: ")
        for tabla in sql_tablename_patterns_dict['read_seqs']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                counts = self.execute_select_query(
                    "SELECT COUNT(*) FROM {tablename};".format(tablename=tablename))
                print(tablename, " : ", counts[0][0])

        print("=== Genomes References tables igor-genomes: ")
        for tabla in sql_tablename_patterns_dict['ref_genome']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                counts = self.execute_select_query(
                    "SELECT COUNT(*) FROM {tablename};".format(tablename=tablename))
                print(tablename, " : ", counts[0][0])

        print("=== Alignments tables igor-alignments: ")
        for tabla in sql_tablename_patterns_dict['align']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                counts = self.execute_select_query(
                    "SELECT COUNT(*) FROM {tablename};".format(tablename=tablename))
                print(tablename, " : ", counts[0][0])

        print("=== Model tables igor-model: ")
        for tabla in sql_tablename_patterns_dict['model']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                counts = self.execute_select_query(
                    "SELECT COUNT(*) FROM {tablename};".format(tablename=tablename))
                print(tablename, " : ", counts[0][0])

        print("=== Output tables igor-pgen and igor-scenarios: ")
        for tabla in sql_tablename_patterns_dict['output']:
            tables_list = self.get_list_of_tables_with_name(tabla)
            for tablename in tables_list:
                counts = self.execute_select_query(
                    "SELECT COUNT(*) FROM {tablename};".format(tablename=tablename))
                print(tablename, " : ", counts[0][0])


    def copytable_from_source(self, tablename_to_copy, fln_source_db):
        #fln_source_db = igortask.igor_db.flnIgorDB
        sql_ct_from_source_db = """
                        ATTACH '{fln_source_db}' AS sourcedb;
                        INSERT INTO {tablename_to_copy} SELECT * FROM sourcedb.{tablename_to_copy};
                        DETACH sourcedb
                        """
        sql_cmd = sql_ct_from_source_db.format(fln_source_db=fln_source_db, tablename_to_copy=tablename_to_copy)
        self.execute_query(sql_cmd)

    # FIXME: FINISH IT I'M HERE
    def attach_table_from_db(self, flnIgorPgen_to_attach):
        # 1. Attach database flnIgorPgen_to_attach to current connection
        # 2. Create table in flnIgorPgen_to_attach
        # 3. Make a select in base table and insert to new table in flnIgorPgen_to_attach
        # 4. Dettach database flnIgorPgen_to_attach
        """
        ATTACH 'ahorasi.db' as otherdb;
        DETACH otherdb;
        """
        tablename_sql_dict = dict()
        sqlcmd_ct_from_otherdb_template = """
                    SELECT name, sql FROM otherdb.sqlite_master
                    WHERE type ='table' AND name NOT LIKE 'sqlite_%'
                    AND name = '{tablename}';
                    """
        sqlcmd_ct_from_otherdb_template = """
                    SELECT name, sql FROM sqlite_master
                    WHERE type ='table' AND name NOT LIKE 'sqlite_%'
                    AND name = '{tablename}';
                    """


        category = 'sequences'
        if category == 'sequences':
            sql_tablename_patterns_dict['read_seqs']
        if category == 'ref_genome':
            sql_tablename_patterns_dict['ref_genome']
        if category == 'align':
            sql_tablename_patterns_dict['align']
        if category == 'model':
            sql_tablename_patterns_dict['model']

        print( sql_tablename_patterns_dict.keys())
        self.get_dict_of_Igortablename_sql()
        sqlcmd = sqlcmd_ct_from_otherdb_template.format(tablename='IgorPgen')
        record = self.execute_select_query_fetchone(sqlcmd)
        print(record[0])
        print(record[1])


    # def rm_sequences_from_db(self):
    #     for tabla in sql_tablename_patterns_dict['read_seqs']:
    #         tables_list = self.get_list_of_tables_with_name(tabla)
    #         for tablename in tables_list:
    #             sql_cmd_drop = "DROP TABLE IF EXISTS {tablename};".format(tablename=tablename)
    #             self.execute_query(sql_cmd_drop)













"""
J gene with allele.
If referring to a known reference sequence
in a database the relevant gene/allele nomenclature
should be followed (e.g., IGHJ4*02 if using IMGT/GENE-DB).
"""
import collections

class AIRR_gene:
    def __init__(self, gene_type):
        self.gene_type = gene_type # v, d, d2, j
        self.call = None
        self.score = None
        self.identity = None
        self.support = None
        self.cigar = None

        self.sequence_start = None
        self.sequence_end = None
        self.germline_start = None
        self.germline_end = None
        self.alignment_start = None
        self.alignment_end = None

    def to_dict(self):
        dicto = collections.OrderedDict()
        ptype = self.gene_type+"_"
        dicto[ptype + "call"] = self.call
        dicto[ptype + "score"] = self.score
        dicto[ptype + "identity"] = self.identity
        dicto[ptype + "support"] = self.support
        dicto[ptype + "cigar"] = self.cigar

        dicto[ptype + "sequence_start"] = self.sequence_start
        dicto[ptype + "sequence_end"] = self.sequence_end
        dicto[ptype + "germline_start"] = self.germline_start
        dicto[ptype + "germline_end"] = self.germline_end
        dicto[ptype + "alignment_start"] = self.alignment_start
        dicto[ptype + "alignment_end"] = self.alignment_end

        return dicto


class AIRR_VDJ_rearrangement:
    def __init__(self, sequence_id=None, sequence=None, scenario_rank=None, scenario_proba_cond_seq=None, pgen=None):
        self.sequence_id = sequence_id
        self.sequence = sequence
        self.scenario_rank = scenario_rank
        self.scenario_proba_cond_seq = scenario_proba_cond_seq
        self.pgen = pgen

        self.quality = None
        self.sequence_aa = None
        self.rev_comp = None
        self.productive = None
        self.vj_in_frame = None
        self.stop_codon = None
        self.complete_vdj = None
        self.locus = None

        #genes data
        self.v_data = AIRR_gene("v")
        self.d_data = AIRR_gene("d")
        self.j_data = AIRR_gene("j")


        self.sequence_alignment = None
        self.quality_alignment = None
        self.sequence_alignment_aa = None

        self.junction = None
        self.junction_aa = None

        self.n1_length = None

        self.np1 = None
        self.np1_aa = None
        self.np1_length = None

        self.n2_length = None

        self.np2 = None
        self.np2_aa = None
        self.np2_length = None

        self.p3v_length = None
        self.p5d_length = None
        self.p3d_length = None
        self.p5j_length = None

    def to_dict(self):
        dicto = collections.OrderedDict()
        dicto["sequence_id"] = self.sequence_id
        dicto["sequence"] = self.sequence
        dicto["scenario_rank"] = self.scenario_rank
        dicto["scenario_proba_cond_seq"] = self.scenario_proba_cond_seq
        dicto["pgen"] = self.pgen

        dicto.update(self.v_data.to_dict())
        dicto.update(self.d_data.to_dict())
        dicto.update(self.j_data.to_dict())

        dicto["quality"] = self.quality
        dicto["sequence_aa"] = self.sequence_aa
        dicto["rev_comp"] = self.rev_comp
        dicto["productive"] = self.productive
        dicto["vj_in_frame"] = self.vj_in_frame
        dicto["stop_codon"] = self.stop_codon
        dicto["complete_vdj"] = self.complete_vdj
        dicto["locus"] = self.locus
        dicto["sequence_alignment"] = self.sequence_alignment
        dicto["quality_alignment"] = self.quality_alignment
        dicto["sequence_alignment_aa"] = self.sequence_alignment_aa
        dicto["junction"] = self.junction
        dicto["junction_aa"] = self.junction_aa

        dicto["n1_length"] = self.n1_length

        dicto["np1"] = self.np1
        dicto["np1_aa"] = self.np1_aa
        dicto["np1_length"] = self.np1_length

        dicto["n2_length"] = self.n2_length

        dicto["np2"] = self.np2
        dicto["np2_aa"] = self.np2_aa
        dicto["np2_length"] = self.np2_length

        dicto["p3v_length"] = self.p3v_length
        dicto["p5d_length"] = self.p5d_length
        dicto["p3d_length"] = self.p3d_length
        dicto["p5j_length"] = self.p5j_length

        return dicto

    @staticmethod
    def list_of_fields():
        cls = AIRR_VDJ_rearrangement()
        return list(cls.to_dict().keys())



